# flake8: noqa
from .asset_app import AssetApp
from .datasource_app import DataSourceApp
from .project_app import ProjectApp
from .job_app import JobApp
from .model_app import ModelApp
from .analysis_app import AnalysisModuleApp
from .clip_app import VideoClipApp
from .field_app import CustomFieldApp



from .base import BaseEntity

__all__ = [
    'AnalysisModule'
]


class AnalysisModule(BaseEntity):
    """
    An AnalysisModule describes a type of ML process that gets applied to an asset.

    """
    def __init__(self, data):
        super(AnalysisModule, self).__init__(data)

    @property
    def name(self):
        """The name of the AnalysisModule"""
        return self._data['name']

    @property
    def type(self):
        """The type of ML operation the AnalysisModule accomplishes."""
        return self._data['type']

    @property
    def supported_media(self):
        """The types of media supported by the AnalysisModule """
        return self._data['supportedMedia']

    @property
    def category(self):
        """The category/brand of AnalysisModule, example: Google Video Intelligence"""
        return self._data['category']

    @property
    def provider(self):
        """The provider of the AnalysisModule, example as Zorroa, Google, Amazon"""
        return self._data['provider']

    @property
    def description(self):
        """The description of the AnalysisModule"""
        return self._data['description']



from enum import Enum

from .base import BaseEntity
from ..util import as_id

__all__ = [
    'Model',
    'ModelType',
    'Label',
    'LabelScope',
    'ModelTypeInfo'
]


class ModelType(Enum):
    """
    Types of models that can be Trained.
    """

    ZVI_KNN_CLASSIFIER = 0
    """A KMeans clustering model for quickly clustering assets into general groups."""

    ZVI_LABEL_DETECTION = 1
    """Retrain the ResNet50 convolutional neural network with your own labels."""

    ZVI_FACE_RECOGNITION = 2
    """Face Recognition model using a KNN classifier."""

    GCP_LABEL_DETECTION = 4
    """Train a Google AutoML vision model."""

    TF2_IMAGE_CLASSIFIER = 5
    """Provide your own custom Tensorflow2/Keras model"""

    PYTORCH_IMAGE_CLASSIFIER = 5
    """Provide your own custom Pytorch model"""


class LabelScope(Enum):
    """
    Types of label scopes
    """
    TRAIN = 1
    """The label marks the Asset as part of the Training set."""

    TEST = 2
    """The label marks the Asset as part of the Test set."""


class Model(BaseEntity):

    def __init__(self, data):
        super(Model, self).__init__(data)

    @property
    def name(self):
        """The name of the Model"""
        return self._data['name']

    @property
    def module_name(self):
        """The name of the Pipeline Module"""
        return self._data['moduleName']

    @property
    def namespace(self):
        """The name of the Pipeline Module"""
        return 'analysis.{}'.format(self._data['moduleName'])

    @property
    def type(self):
        """The type of model"""
        return ModelType[self._data['type']]

    @property
    def file_id(self):
        """The file ID of the trained model"""
        return self._data['fileId']

    @property
    def ready(self):
        """
        True if the model is fully trained and ready to use.
        Adding new labels will set ready to false.
        """
        return self._data['ready']

    def make_label(self, label, bbox=None, simhash=None, scope=None):
        """
        Make an instance of a Label which can be used to label assets.

        Args:
            label (str): The label name.
            bbox (list[float]): A open bounding box.
            simhash (str): An associated simhash, if any.
            scope (LabelScope): The scope of the image, can be TEST or TRAIN.
                Defaults to TRAIN.
        Returns:
            Label: The new label.
        """
        return Label(self, label, bbox=bbox, simhash=simhash, scope=scope)

    def make_label_from_prediction(self, label, prediction, scope=None):
        """
        Make a label from a prediction.  This will copy the bbox
        and simhash from the prediction, if any.

        Args:
            label (str): A name for the prediction.
            prediction (dict): A prediction from an analysis namespace.s
            scope (LabelScope): The scope of the image, can be TEST or TRAIN.
                Defaults to TRAIN.
        Returns:
            Label: A new label
        """
        return Label(self, label,
                     bbox=prediction.get('bbox'),
                     simhash=prediction.get('simhash'),
                     scope=scope)

    def get_label_search(self, scope=None):
        """
        Return a search that can be used to query all assets
        with labels.

        Args:
            scope (LabelScope): An optional label scope to filter by.

        Returns:
            dict: A search to pass to an asset search.
        """
        search = {
            'size': 64,
            'sort': [
                '_doc'
            ],
            '_source': ['labels', 'files'],
            'query': {
                'nested': {
                    'path': 'labels',
                    'query': {
                        'bool': {
                            'must': [
                                {'term': {'labels.modelId': self.id}}
                            ]
                        }
                    }
                }
            }
        }

        if scope:
            must = search['query']['nested']['query']['bool']['must']
            must.append({'term': {'labels.scope': scope.name}})

        return search

    def get_confusion_matrix_search(self, min_score=0.0, max_score=1.0, test_set_only=True):
        """
        Returns a search query with aggregations that can be used to create a confusion
        matrix.

        Args:
            min_score (float): Minimum confidence score to return results for.
            max_score (float): Maximum confidence score to return results for.
            test_set_only (bool): If True only assets with TEST labels will be evaluated.

        Returns:
            dict: A search to pass to an asset search.

        """
        prediction_term_map = {
            ModelType.ZVI_KNN_CLASSIFIER: f'{self.namespace}.label',
            ModelType.ZVI_FACE_RECOGNITION: f'{self.namespace}.predictions.label'
        }
        score_map = {ModelType.ZVI_KNN_CLASSIFIER: f'{self.namespace}.score',
                     ModelType.ZVI_LABEL_DETECTION: f'{self.namespace}.score',
                     ModelType.ZVI_FACE_RECOGNITION: f'{self.namespace}.predictions.score'}
        if self.type not in prediction_term_map:
            raise TypeError(f'Cannot create a confusion matrix search for {self.type} models.')
        search_query = {
            "size": 0,
            "query": {
                "bool": {
                    "filter": [
                        {"range": {score_map[self.type]: {"gte": min_score, "lte": max_score}}}
                    ]
                }
            },
            "aggs": {
                "nested_labels": {
                    "nested": {
                        "path": "labels"
                    },
                    "aggs": {
                        "model_train_labels": {
                            "filter": {
                                "bool": {
                                    "must": [
                                        {"term": {"labels.modelId": self.id}}
                                    ]
                                }
                            },
                            "aggs": {
                                "labels": {
                                    "terms": {"field": "labels.label"},
                                    "aggs": {
                                        "predictions_by_label": {
                                            "reverse_nested": {},
                                            "aggs": {
                                                "predictions": {
                                                    "terms": {
                                                        "field": prediction_term_map[self.type]
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        if test_set_only:
            (search_query
             ['aggs']
             ['nested_labels']
             ['aggs']
             ['model_train_labels']
             ['filter']
             ['bool']
             ['must'].append({"term": {"labels.scope": "TEST"}}))
        return search_query


class ModelTypeInfo:
    """
    Additional properties related to each ModelType.
    """
    def __init__(self, data):
        self._data = data

    @property
    def name(self):
        """The name of the model type."""
        return self._data['name']

    @property
    def description(self):
        """The description of the model type."""
        return self._data['description']

    @property
    def objective(self):
        """The objective of the model, LABEL_DETECTION, FACE_RECOGNITION, etc"""
        return self._data['objective']

    @property
    def provider(self):
        """The company that maintains the structure and algorithm for the model."""
        return self._data['provider']

    @property
    def min_concepts(self):
        """The minimum number of unique concepts a model must have before it can be trained."""
        return self._data['minConcepts']

    @property
    def min_examples(self):
        """
        The minimum number of examples per concept a
        model must have before it can be trained.
        """
        return self._data['minExamples']


class Label:
    """
    A Label that can be added to an Asset either at import time
    or once the Asset has been imported.
    """

    def __init__(self, model, label, bbox=None, simhash=None, scope=None):
        """
        Create a new label.

        Args:
            model: (Model): The model the label is for.
            label (str): The label itself.
            bbox (list): A optional list of floats for a bounding box.
            simhash (str): An optional similatity hash.
            scope (LabelScope): The scope of the image, can be TEST or TRAIN.
                Defaults to TRAIN.
        """
        self.model_id = as_id(model)
        self.label = label
        self.bbox = bbox
        self.simhash = simhash
        self.scope = scope or LabelScope.TRAIN

    def for_json(self):
        """Returns a dictionary suitable for JSON encoding.

        The ZpsJsonEncoder will call this method automatically.

        Returns:
            :obj:`dict`: A JSON serializable version of this Document.

        """
        return {
            'modelId': self.model_id,
            'label': self.label,
            'bbox': self.bbox,
            'simhash': self.simhash,
            'scope': self.scope.name
        }



import io
import os
import requests
from collections import namedtuple

from ..entity import Asset, StoredFile, FileUpload, FileTypes, Job, VideoClip
from ..search import AssetSearchResult, AssetSearchScroller, SimilarityQuery, SearchScroller
from ..util import as_collection, as_id_collection, as_id


class AssetApp(object):

    def __init__(self, app):
        self.app = app

    def batch_import_files(self, files, modules=None):
        """
        Import a list of FileImport instances.

        Args:
            files (list of FileImport): The list of files to import as Assets.
            modules (list): A list of Pipeline Modules to apply to the data.

        Notes:
            Example return value:
                {
                  "bulkResponse" : {
                    "took" : 15,
                    "errors" : false,
                    "items" : [ {
                      "create" : {
                        "_index" : "yvqg1901zmu5bw9q",
                        "_type" : "_doc",
                        "_id" : "dd0KZtqyec48n1q1fniqVMV5yllhRRGx",
                        "_version" : 1,
                        "result" : "created",
                        "forced_refresh" : true,
                        "_shards" : {
                          "total" : 1,
                          "successful" : 1,
                          "failed" : 0
                        },
                        "_seq_no" : 0,
                        "_primary_term" : 1,
                        "status" : 201
                      }
                    } ]
                  },
                  "failed" : [ ],
                  "created" : [ "dd0KZtqyec48n1q1fniqVMV5yllhRRGx" ],
                  "jobId" : "ba310246-1f87-1ece-b67c-be3f79a80d11"
                }

        Returns:
            dict: A dictionary containing an ES bulk response, failed files,
            and created asset ids.

        """
        body = {
            "assets": files,
            "modules": modules
        }
        return self.app.client.post("/api/v3/assets/_batch_create", body)

    def batch_upload_files(self, files, modules=None):
        """
        Batch upload a list of files and return a structure which contains
        an ES bulk response object, a list of failed file paths, a list of created
        asset Ids, and a processing jobId.

        Args:
            files (list of FileUpload):
            modules (list): A list of Pipeline Modules to apply to the data.

        Notes:
            Example return value:
                {
                  "bulkResponse" : {
                    "took" : 15,
                    "errors" : false,
                    "items" : [ {
                      "create" : {
                        "_index" : "yvqg1901zmu5bw9q",
                        "_type" : "_doc",
                        "_id" : "dd0KZtqyec48n1q1fniqVMV5yllhRRGx",
                        "_version" : 1,
                        "result" : "created",
                        "forced_refresh" : true,
                        "_shards" : {
                          "total" : 1,
                          "successful" : 1,
                          "failed" : 0
                        },
                        "_seq_no" : 0,
                        "_primary_term" : 1,
                        "status" : 201
                      }
                    } ]
                  },
                  "failed" : [ ],
                  "created" : [ "dd0KZtqyec48n1q1fniqVMV5yllhRRGx" ],
                  "jobId" : "ba310246-1f87-1ece-b67c-be3f79a80d11"
                }

        Returns:
            dict: A dictionary containing an ES bulk response, failed files,
            and created asset ids.
        """
        files = as_collection(files)
        file_paths = [f.uri for f in files]
        body = {
            "assets": files,
            "modules": modules
        }
        return self.app.client.upload_files("/api/v3/assets/_batch_upload",
                                            file_paths, body)

    def batch_upload_directory(self, path, file_types=None,
                               batch_size=50, modules=None, callback=None):
        """
        Recursively upload all files in the given directory path.

        This method takes an optional callback function which takes two
        arguments, files and response.  This callback is called for
        each batch of files submitted.

        Examples:

            def batch_callback(files, response):
                print("--processed files--")
                for path in files:
                    print(path)
                print("--zvi response--")
                pprint.pprint(rsp)

            app.assets.batch_upload_directory("/home", file_types=['images'],
                callback=batch_callback)

        Args:
            path (str): A file path to a directory.
            file_types (list): a list of file extensions and/or
                categories(documents, images, videos)
            batch_size (int) The number of files to upload per batch.
            modules (list): An array of modules to apply to the files.
            callback (func): A function to call for every batch

        Returns:
            dict: A dictionary containing batch operation counters.
        """
        batch = []
        totals = {
            "file_count": 0,
            "file_size": 0,
            "batch_count": 0,
        }

        def process_batch():
            totals['batch_count'] += 1
            totals['file_count'] += len(batch)
            totals['file_size'] += sum([os.path.getsize(f) for f in batch])

            rsp = self.batch_upload_files(
                [FileUpload(f) for f in batch], modules)
            if callback:
                callback(batch.copy(), rsp)
            batch.clear()

        file_types = FileTypes.resolve(file_types)
        for root, dirs, files in os.walk(path):
            for fname in files:
                if fname.startswith("."):
                    continue
                _, ext = os.path.splitext(fname)
                if not ext:
                    continue
                if ext[1:].lower() not in file_types:
                    continue
                batch.append(os.path.abspath(os.path.join(root, fname)))
                if len(batch) >= batch_size:
                    process_batch()

        if batch:
            process_batch()

        return totals

    def delete_asset(self, asset):
        """
        Delete the given asset.

        Args:
            asset (mixed): unique Id or Asset instance.

        Returns:
            bool: True if the asset was deleted.

        """
        asset_id = as_id(asset)
        return self.app.client.delete("/api/v3/assets/{}".format(asset_id))['success']

    def batch_delete_assets(self, assets):
        """
        Batch delete the given list of Assets or asset ids.

        Args:
            assets (list): A list of Assets or unique asset ids.

        Returns:
            dict: A dictionary containing deleted and errored asset Ids.
        """
        body = {
            "assetIds": as_id_collection(assets)
        }
        return self.app.client.delete("/api/v3/assets/_batch_delete", body)

    def search(self, search=None, fetch_source=True):
        """
        Perform an asset search using the ElasticSearch query DSL.

        See Also:
            For search/query format.
            https://www.elastic.co/guide/en/elasticsearch/reference/6.4/search-request-body.html

        Args:
            search (dict): The ElasticSearch search to execute.
            fetch_source: (bool): If true, the full JSON document for each asset is returned.
        Returns:
            AssetSearchResult - an AssetSearchResult instance.
        """
        if not fetch_source:
            search['_source'] = False
        return AssetSearchResult(self.app, search)

    def scroll_search(self, search=None, timeout="1m"):
        """
        Perform an asset scrolled search using the ElasticSearch query DSL.

        See Also:
            For search/query format.
            https://www.elastic.co/guide/en/elasticsearch/reference/6.4/search-request-body.html

        Args:
            search (dict): The ElasticSearch search to execute
            timeout (str): The scroll timeout.  Defaults to 1 minute.
        Returns:
            AssetSearchScroll - an AssetSearchScroller instance which is a generator
                by nature.

        """
        return AssetSearchScroller(self.app, search, timeout)

    def reprocess_search(self, search, modules):
        """
        Reprocess the given search with the supplied modules.

        Args:
            search (dict): An ElasticSearch search.
            modules (list): A list of module names to apply.

        Returns:
            dict: Contains a Job and the number of assets to be processed.
        """
        body = {
            "search": search,
            "modules": modules
        }
        rsp = self.app.client.post("/api/v3/assets/_search/reprocess", body)
        return ReprocessSearchResponse(rsp["assetCount"], Job(rsp["job"]))

    def scroll_search_clips(self, asset, search=None, timeout="1m"):
        """
        Scroll through clips for given asset using the ElasticSearch query DSL.

        Args:
            asset (Asset): The asset or unique AssetId.
            search (dict): The ElasticSearch search to execute
            timeout (str): The scroll timeout.  Defaults to 1 minute.

        Returns:
            SearchScroller  a clip scroller instance for generating VideoClips.

        """
        asset_id = as_id(asset)
        return SearchScroller(
            VideoClip, f'/api/v3/assets/{asset_id}/clips/_search', self.app, search, timeout
        )

    def reprocess_assets(self, assets, modules):
        """
        Reprocess the given array of assets with the given modules.

        Args:
            assets (list): A list of Assets or asset unique Ids.
            modules (list): A list of Pipeline module names or ides.

        Returns:
            Job: The job responsible for processing the assets.
        """
        asset_ids = [getattr(asset, "id", asset) for asset in as_collection(assets)]
        body = {
            "search": {
                "query": {
                    "terms": {
                        "_id": asset_ids
                    }
                }
            },
            "modules": as_collection(modules)
        }

        return self.app.client.post("/api/v3/assets/_search/reprocess", body)

    def get_asset(self, id):
        """
        Return the asset with the given unique Id.

        Args:
            id (str): The unique ID of the asset.

        Returns:
            Asset: The Asset
        """
        return Asset(self.app.client.get("/api/v3/assets/{}".format(id)))

    def update_labels(self, assets, add_labels=None, remove_labels=None):
        """
        Update the Labels on the given array of assets.

        Args:
            assets (mixed): An Asset, asset ID, or a list of either type.
            add_labels (list[Label]): A Label or list of Label to add.
            remove_labels (list[Label]): A Label or list of Label to remove.
        Returns:
            dict: An request status dict

        """
        ids = as_id_collection(assets)
        body = {}
        if add_labels:
            body['add'] = dict([(a, as_collection(add_labels)) for a in ids])
        if remove_labels:
            body['remove'] = dict([(a, as_collection(remove_labels)) for a in ids])
        if not body:
            raise ValueError("Must pass at least and add_labels or remove_labels argument")
        return self.app.client.put("/api/v3/assets/_batch_update_labels", body)

    def update_custom_fields(self, asset, values):
        """
        Set the values of custom metadata fields.

        Args:
            asset (Asset): The asset or unique Asset id.
            values (dict): A dictionary of values.

        Returns:
            dict: A status dictionary with failures or succcess
        """
        body = {
            "update": {
                as_id(asset): values
            }
        }
        return self.app.client.put("/api/v3/assets/_batch_update_custom_fields", body)

    def batch_update_custom_fields(self, update):
        """
        Set the values of custom metadata fields.

        Examples:
            {
                "asset-id1": {"shoe": "nike"},
                "asset-id2": {"country": "New Zealand"}
            }

        Args:
            update (dict): A dict o dicts which describe the
        Returns:
            dict: A status dictionary with failures or success
        """
        body = {
            'update': update
        }
        return self.app.client.put('/api/v3/assets/_batch_update_custom_fields', body)

    def download_file(self, stored_file, dst_file=None):
        """
        Download given file and store results in memory, or optionally
        a destination file.  The stored_file ID can be specified as
        either a string like "assets/<id>/proxy/image_450x360.jpg"
        or a StoredFile instance can be used.

        Args:
            stored_file (mixed): The StoredFile instance or its ID.
            dst_file (str): An optional destination file path.

        Returns:
            io.BytesIO instance containing the binary data or if
                a destination path was provided the size of the
                file is returned.

        """
        if isinstance(stored_file, str):
            path = stored_file
        elif isinstance(stored_file, StoredFile):
            path = stored_file.id
        else:
            raise ValueError("stored_file must be a string or StoredFile instance")

        rsp = self.app.client.get("/api/v3/files/_stream/{}".format(path), is_json=False)
        if dst_file:
            with open(dst_file, 'wb') as fp:
                fp.write(rsp.content)
            return os.path.getsize(dst_file)
        else:
            return io.BytesIO(rsp.content)

    def stream_file(self, stored_file, chunk_size=1024):
        """
        Streams a file by iteratively returning chunks of the file using a generator. This
        can be useful when developing web applications and a full download of the file
        before continuing is not necessary.

        Args:
            stored_file (mixed): The StoredFile instance or its ID.
            chunk_size (int): The byte sizes of each requesting chunk. Defaults to 1024.

        Yields:
            generator (File-like Object): Content of the file.

        """
        if isinstance(stored_file, str):
            path = stored_file
        elif isinstance(stored_file, StoredFile):
            path = stored_file.id
        else:
            raise ValueError("stored_file must be a string or StoredFile instance")

        url = self.app.client.get_url('/api/v3/files/_stream/{}'.format(path))
        response = requests.get(url, verify=self.app.client.verify,
                                headers=self.app.client.headers(), stream=True)

        for block in response.iter_content(chunk_size):
            yield block

    def get_sim_hashes(self, images):
        """
        Return a similarity hash for the given array of images.

        Args:
            images (mixed): Can be an file handle (opened with 'rb'), or
                path to a file.
        Returns:
            list of str: A list of similarity hashes.

        """
        return self.app.client.upload_files("/ml/v1/sim-hash",
                                            as_collection(images), body=None)

    def get_sim_query(self, images, min_score=0.75):
        """
        Analyze the given image files and return a SimilarityQuery which
        can be used in a search.

        Args:
            images (mixed): Can be an file handle (opened with 'rb'), or
                path to a file.
            min_score (float): A float between, the higher the value the more similar
                the results.  Defaults to 0.75

        Returns:
            SimilarityQuery: A configured SimilarityQuery
        """
        return SimilarityQuery(self.get_sim_hashes(images), min_score)


"""
A named tuple to define a ReprocessSearchResponse
"""
ReprocessSearchResponse = namedtuple('ReprocessSearchResponse', ["asset_count", "job"])



import json
import logging
import os

from ..client import to_json
from ..util import as_collection

__all__ = [
    'Asset',
    'FileImport',
    'FileUpload',
    'StoredFile',
    'FileTypes'
]

logger = logging.getLogger(__name__)


class DocumentMixin(object):
    """
    A Mixin class which provides easy access to a deeply nested dictionary.
    """

    def __init__(self):
        self.document = {}

    def set_attr(self, attr, value):
        """Set the value of an attribute.

        Args:
            attr (str): The attribute name in dot notation format.
                ex: 'foo.bar'
            value (:obj:`object`): value: The value for the particular
                attribute. Can be any json serializable type.
        """
        self.__set_attr(attr, value)

    def del_attr(self, attr):
        """
        Delete the attribute from the document.  If the attribute does not exist
        or is protected by a manual field edit then return false.  Otherwise,
        delete the attribute and return true.

        Args:
            attr (str): The attribute name.

        Returns:
            bool: True if the attribute was deleted.

        """
        doc = self.document
        parts = attr.split(".")
        for k in parts[0:-1]:
            if not isinstance(doc, dict) or k not in doc:
                return False
            doc = doc.get(k)

        attr_name = parts[-1]
        try:
            del doc[attr_name]
            return not self.attr_exists(attr)
        except KeyError:
            return False

    def get_attr(self, attr, default=None):
        """Get the given attribute to the specified value.

        Args:
            attr (str): The attribute name in dot notation format.
                ex: 'foo.bar'
            default (:obj:`mixed`) The default value if no attr exists.

        Returns:
            mixed: The value of the attribute.

        """
        doc = self.document
        parts = attr.split(".")
        for k in parts:
            if not isinstance(doc, dict) or k not in doc:
                return default
            doc = doc.get(k)
        return doc

    def attr_exists(self, attr):
        """
        Return true if the given attribute exists.

        Args:
            attr (str): The name of the attribute to check.

        Returns:
            bool: true if the attr exists.

        """
        doc = self.document
        parts = attr.split(".")
        for k in parts[0:len(parts) - 1]:
            if k not in doc:
                return False
            doc = doc.get(k)
        return parts[-1] in doc

    def add_analysis(self, name, val):
        """Add an analysis structure to the document.

        Args:
            name (str): The name of the analysis
            val (mixed): the value/result of the analysis.

        """
        if not name:
            raise ValueError("Analysis requires a unique name")
        attr = "analysis.%s" % name
        if val is None:
            self.set_attr(attr, None)
        else:
            self.set_attr(attr, json.loads(to_json(val)))

    def get_analysis(self, namespace):
        """
        Return the the given analysis data under the the given name.

        Args:
            namespace (str): The  model namespace / pipeline module name.

        Returns:
            dict: An arbitrary dictionary containing predictions, content, etc.

        """
        name = getattr(namespace, "namespace", "analysis.{}".format(namespace))
        return self.get_attr(name)

    def get_predicted_labels(self, namespace, min_score=None):
        """
        Get all predictions made by the given label prediction module. If no
        label predictions are present, returns None.

        Args:
            namespace (str): The analysis namespace, example 'zvi-label-detection'.
            min_score (float): Filter results by a minimum score.

        Returns:
            list: A list of dictionaries containing the predictions

        """
        name = getattr(namespace, "namespace", "analysis.{}".format(namespace))
        predictions = self.get_attr(f'{name}.predictions')
        if not predictions:
            return None
        if min_score:
            return [pred for pred in predictions if pred['score'] >= min_score]
        else:
            return predictions

    def get_predicted_label(self, namespace, label):
        """
        Get a prediction made by the given label prediction module.  If no
        label predictions are present, returns None.

        Args:
            namespace (str): The model / module name that created the prediction.
            label (mixed): A label name or integer index of a prediction.

        Returns:
            dict: a prediction dict with a label, score, etc.
        """

        preds = self.get_predicted_labels(namespace)
        if not preds:
            return None

        if isinstance(label, str):
            preds = [pred for pred in preds if pred['label'] == label]
            label = 0

        try:
            return preds[label]
        except IndexError:
            return None

    def extend_list_attr(self, attr, items):
        """
        Adds the given items to the given attr. The attr must be a list or set.

        Args:
            attr (str): The name of the attribute
            items (:obj:`list` of :obj:`mixed`): A list of new elements.

        """
        items = as_collection(items)
        all_items = self.get_attr(attr)
        if all_items is None:
            all_items = set()
            self.set_attr(attr, all_items)
        try:
            all_items.update(items)
        except AttributeError:
            all_items.extend(items)

    def __set_attr(self, attr, value):
        """
        Handles setting an attribute value.

        Args:
            attr (str): The attribute name in dot notation format.  ex: 'foo.bar'
            value (mixed): The value for the particular attribute.
                Can be any json serializable type.
        """
        doc = self.document
        parts = attr.split(".")
        for k in parts[0:len(parts) - 1]:
            if k not in doc:
                doc[k] = {}
            doc = doc[k]
        if isinstance(value, dict):
            doc[parts[-1]] = value
        else:
            try:
                doc[parts[-1]] = value.for_json()
            except AttributeError:
                doc[parts[-1]] = value

    def __setitem__(self, field, value):
        self.set_attr(field, value)

    def __getitem__(self, field):
        return self.get_attr(field)


class FileImport(object):
    """
    An FileImport is used to import a new file and metadata into ZMLP.
    """

    def __init__(self, uri, custom=None, page=None, label=None, tmp=None):
        """
        Construct an FileImport instance which can point to a remote URI.

        Args:
            uri (str): a URI locator to the file asset.
            custom (dict): Values for custom metadata fields.
            page (int): The specific page to import if any.
            label (Label): An optional Label which will add the file to
                a Model training set.
            tmp: (dict): A dict of temp attrs that are removed after procssing.
        """
        super(FileImport, self).__init__()
        self.uri = uri
        self.custom = custom or {}
        self.page = page
        self.label = label
        self.tmp = tmp

    def for_json(self):
        """Returns a dictionary suitable for JSON encoding.

        The ZpsJsonEncoder will call this method automatically.

        Returns:
            :obj:`dict`: A JSON serializable version of this Document.

        """
        return {
            "uri": self.uri,
            "custom": self.custom,
            "page": self.page,
            "label": self.label,
            "tmp": self.tmp
        }

    def __setitem__(self, field, value):
        self.custom[field] = value

    def __getitem__(self, field):
        return self.custom[field]


class FileUpload(FileImport):
    """
    FileUpload instances point to a local file that will be uploaded for analysis.
    """

    def __init__(self, path, custom=None, page=None, label=None):
        """
        Create a new FileUpload instance.

        Args:
            path (str): A path to a file, the file must exist.
            custom (dict): Values for pre-created custom metadata fields.
            page (int): The specific page to import if any.
            label (Label): An optional Label which will add the file to
                a Model training set.
        """
        super(FileUpload, self).__init__(
            os.path.normpath(os.path.abspath(path)), custom, page, label)

        if not os.path.exists(path):
            raise ValueError('The path "{}" does not exist'.format(path))

    def for_json(self):
        """Returns a dictionary suitable for JSON encoding.

        The ZpsJsonEncoder will call this method automatically.

        Returns:
            :obj:`dict`: A JSON serializable version of this Document.

        """
        return {
            "uri": self.uri,
            "page": self.page,
            "label": self.label,
            "custom": self.custom
        }


class Asset(DocumentMixin):
    """
    An Asset represents a single processed file.  Assets start out
    in the 'CREATED' state, which indicates they've been created by not processed.
    Once an asset has been processed and augmented with files created by various
    analysis modules, the Asset will move into the 'ANALYZED' state.
    """

    def __init__(self, data):
        super(Asset, self).__init__()
        if not data:
            raise ValueError("Error creating Asset instance, Assets must have an id.")
        self.id = data.get("id")
        self.document = data.get("document", {})
        self.score = data.get("score", 0)
        self.inner_hits = data.get("inner_hits", [])

    @staticmethod
    def from_hit(hit):
        """
        Converts an ElasticSearch hit into an Asset.

        Args:
            hit (dict): An raw ES document

        Returns:
            Asset: The Asset.
        """
        return Asset({
            'id': hit['_id'],
            'score': hit.get('_score', 0),
            'document': hit.get('_source', {}),
            'inner_hits': hit.get('inner_hits', [])})

    @property
    def uri(self):
        """
        The URI of the asset.

        Returns:
            str: The URI of the data.

        """
        return self.get_attr("source.path")

    @property
    def extension(self):
        """
        The file extension of the asset, lower cases.

        Returns:
            str: The file extension

        """
        return self.get_attr("source.extension").lower()

    def add_file(self, stored_file):
        """
        Adds the StoredFile record to the asset's list of associated files.

        Args:
            stored_file (StoredFile): A file that has been stored in ZMLP

        Returns:
            bool: True if the file was added to the list, False if it was a duplicate.

        """
        # Ensure the file doesn't already exist in the metadata
        if not self.get_files(id=stored_file.id):
            files = self.get_attr("files") or []
            files.append(stored_file._data)
            self.set_attr("files", files)
            return True
        return False

    def get_files(self, name=None, category=None, mimetype=None, extension=None,
                  id=None, attrs=None, attr_keys=None, sort_func=None):
        """
        Return all stored files associated with this asset.  Optionally
        filter the results.

        Args:
            name (str): The associated files name.
            category (str): The associated files category, eg proxy, backup, etc.
            mimetype (str): The mimetype must start with this string.
            extension: (str): The file name must have the given extension.
            attrs (dict): The file must have all of the given attributes.
            attr_keys: (list): A list of attribute keys that must be present.
            sort_func: (func): A lambda function for sorting the result.
        Returns:
            list of StoredFile: A list of ZMLP file records.

        """
        result = []
        files = self.get_attr("files") or []
        for fs in files:
            match = True
            if id and not any((item for item in as_collection(id)
                               if fs["id"] == item)):
                match = False
            if name and not any((item for item in as_collection(name)
                                 if fs["name"] == item)):
                match = False
            if category and not any((item for item in as_collection(category)
                                     if fs["category"] == item)):
                match = False
            if mimetype and not any((item for item in as_collection(mimetype)
                                     if fs["mimetype"].startswith(item))):
                match = False
            if extension and not any((item for item in as_collection(extension)
                                      if fs["name"].endswith("." + item))):
                match = False

            file_attrs = fs.get("attrs", {})
            if attr_keys:
                if not any(key in file_attrs for key in as_collection(attr_keys)):
                    match = False

            if attrs:
                for k, v in attrs.items():
                    if file_attrs.get(k) != v:
                        match = False
            if match:
                result.append(StoredFile(fs))

        if sort_func:
            result = sorted(result, key=sort_func)

        return result

    def get_thumbnail(self, level):
        """
        Return an thumbnail StoredFile record for the Asset. The level
        corresponds size of the thumbnail, 0 for the smallest, and
        up to N for the largest.  Levels 0,1,and 2 are smaller than
        the source media, level 3 or above  (if they exist) will
        be full resolution or higher images used for OCR purposes.

        To download the thumbnail call app.assets.download_file(stored_file)

        Args:
            level (int): The size level, 0 for smallest up to N.

        Returns:
            StoredFile: A StoredFile instance or None if no image proxies exist.
        """
        files = self.get_files(mimetype="image/", category="proxy",
                               sort_func=lambda f: f.attrs.get('width', 0))
        if not files:
            return None
        if level >= len(files):
            level = -1
        return files[level]

    def get_inner_hits(self, name):
        """
        Return any inner hits from a collapse query.

        Args:
            name (str): The inner hit name.

        Returns:
            list[Asset]:  A list of Assets.
        """
        try:
            return [Asset.from_hit(hit) for hit in self.inner_hits[name]['hits']['hits']]
        except KeyError:
            return []

    def for_json(self):
        """Returns a dictionary suitable for JSON encoding.

        The ZpsJsonEncoder will call this method automatically.

        Returns:
            :obj:`dict`: A JSON serializable version of this Document.

        """
        return {
            "id": self.id,
            "uri": self.get_attr("source.path"),
            "document": self.document,
            "page": self.get_attr("media.pageNumber"),
        }

    def __str__(self):
        return "<Asset id='{}'/>".format(self.id)

    def __repr__(self):
        return "<Asset id='{}' at {}/>".format(self.id, hex(id(self)))

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        if not getattr(other, "id"):
            return False
        return other.id == self.id


class StoredFile(object):
    """
    The StoredFile class represents a supporting file that has been stored in ZVI.
    """

    def __init__(self, data):
        self._data = data

    @property
    def id(self):
        """
        The unique ID of the file.
        """
        return self._data['id']

    @property
    def name(self):
        """
        The file name..
        """
        return self._data['name']

    @property
    def category(self):
        """
        The file category.
        """
        return self._data['category']

    @property
    def attrs(self):
        """
        Arbitrary attributes.
        """
        return self._data['attrs']

    @property
    def mimetype(self):
        """
        The file mimetype.
        """
        return self._data['mimetype']

    @property
    def size(self):
        """
        The size of the file.
        """
        return self._data['size']

    @property
    def cache_id(self):
        """
        A string suitable for on-disk caching/filenames.  Replaces
        all slashes in id with underscores.
        """
        return self.id.replace("/", "_")

    def __str__(self):
        return "<StoredFile {}>".format(self.id)

    def __eq__(self, other):
        return other.id

    def __hash__(self):
        return hash(self.id)

    def for_json(self):
        """Return a JSON serialized copy.

        Returns:
            :obj:`dict`: A json serializable dict.
        """
        serializable_dict = {}
        attrs = self._data.keys()
        for attr in attrs:
            if getattr(self, attr, None) is not None:
                serializable_dict[attr] = getattr(self, attr)
        return serializable_dict


class FileTypes:
    """
    A class for storing the supported file types.
    """

    videos = frozenset(['mov', 'mp4', 'mpg', 'mpeg', 'm4v', 'webm', 'ogv', 'ogg', 'mxf', 'avi'])
    """A set of supported video file formats."""

    images = frozenset(["bmp", "cin", "dpx", "gif", "jpg",
                        "jpeg", "exr", "png", "psd", "rla", "tif", "tiff",
                        "dcm", "rla"])
    """A set of supported image file formats."""

    documents = frozenset(['pdf', 'doc', 'docx', 'ppt', 'pptx', 'xls', 'xlsx', 'vsd', 'vsdx'])
    """A set of supported document file formats."""

    all = frozenset(videos.union(images).union(documents))

    """A set of all supported file formats."""

    @classmethod
    def resolve(cls, file_types):
        """
        Resolve a list of file extenions or types (images, documents, videos) to
        a supported list of extensions.

        Args:
            file_types (list): A list of file extensions, dot not included.

        Returns:
            list: The valid list of extensions from the given list

        """
        file_types = as_collection(file_types)
        if not file_types:
            return cls.all
        result = set()
        for file_type in file_types:
            if file_type in cls.all:
                result.add(file_type)
            else:
                exts = getattr(cls, file_type, None)
                if exts:
                    result.update(exts)

        return sorted(list(result))



from .base import BaseEntity

__all__ = [
    'DataSource'
]


class DataSource(BaseEntity):
    """
    A DataSource is a remote source for Assets that can be
    iterated by the Analysis framework and imported
    in a single import Job.
    """

    def __init__(self, data):
        super(DataSource, self).__init__(data)

    @property
    def id(self):
        """The id of the DataSource"""
        return self._data['id']

    @property
    def name(self):
        """The name of the DataSource"""
        return self._data['name']

    @property
    def uri(self):
        """The URI of the DataSource"""
        return self._data['uri']

    @property
    def file_types(self):
        """The file type filter for the DataSource"""
        return self._data.get('file_types', [])

    @property
    def modules(self):
        """The type of modules done to the DataSource"""
        return self._data.get('modules', [])

    @property
    def credentials(self):
        """The type of credentials attached to DataSource"""
        return self._data.get('credentials', [])




from .base import BaseEntity


class CustomField(BaseEntity):
    """
    Fields are used to store your own metadata on an asset.
    """

    def __init__(self, data):
        super(CustomField, self).__init__(data)

    @property
    def type(self):
        """The ES field data type."""
        return self._data['type']

    @property
    def name(self):
        """The base field name."""
        return self._data['name']

    @property
    def es_field_name(self):
        """The full ES field name"""
        return self._data['esField']



from enum import Enum

from .base import BaseEntity

__all__ = [
    'Project',
    'ProjectTier'
]


class ProjectTier(Enum):
    """
    ProjectTiers determine which features are available to a project.s
    """

    ESSENTIALS = 0
    """Allows the use of essentials features."""

    PREMIER = 1
    """Allows the use of premier features."""


class Project(BaseEntity):
    """
    Represents a ZMLP Project.

    """
    def __init__(self, data):
        super(Project, self).__init__(data)

    @property
    def name(self):
        """The project's unique name."""
        return self._data['name']

    @property
    def id(self):
        """The project's unique id."""
        return self._data['id']

    @property
    def tier(self):
        """The project billing tier"""
        return ProjectTier[self._data['tier']]



import base64
import logging
import os

from . import AssetApp, DataSourceApp, ProjectApp, \
    JobApp, ModelApp, AnalysisModuleApp, VideoClipApp, CustomFieldApp
from ..client import ZmlpClient, DEFAULT_SERVER

logger = logging.getLogger(__name__)


class ZmlpApp(object):
    """
    Exposes the main ZMLP API.

    """
    def __init__(self, apikey, server=None):
        """
        Initialize a ZMLP Application instance.

        Args:
            apikey (mixed): An API key, can be either a key or file handle.
            server (str): The URL to the ZMLP API server, defaults cloud api.
        """
        logger.debug("Initializing ZMLP to {}".format(server))
        self.client = ZmlpClient(apikey, server or
                                 os.environ.get("ZMLP_SERVER", DEFAULT_SERVER))
        self.assets = AssetApp(self)
        self.datasource = DataSourceApp(self)
        self.projects = ProjectApp(self)
        self.jobs = JobApp(self)
        self.models = ModelApp(self)
        self.analysis = AnalysisModuleApp(self)
        self.clips = VideoClipApp(self)
        self.fields = CustomFieldApp(self)


def app_from_env():
    """
    Create a ZmlpApp configured via environment variables. This method
    will not throw if the environment is configured improperly, however
    attempting the use the ZmlpApp instance to make a request
    will fail.

    - ZMLP_APIKEY : A base64 encoded API key.
    - ZMLPL_APIKEY_FILE : A path to a JSON formatted API key.
    - ZMLP_SERVER : The URL to the ZMLP API server.

    Returns:
        ZmlpClient : A configured ZmlpClient

    """
    apikey = None
    if 'ZMLP_APIKEY' in os.environ:
        apikey = os.environ['ZMLP_APIKEY']
    elif 'ZMLP_APIKEY_FILE' in os.environ:
        with open(os.environ['ZMLP_APIKEY_FILE'], 'rb') as fp:
            apikey = base64.b64encode(fp.read())
    return ZmlpApp(apikey, os.environ.get('ZMLP_SERVER'))



from ..util import as_id, as_id_collection
from ..search import VideoClipSearchResult, VideoClipSearchScroller
from ..entity import VideoClip


class VideoClipApp:
    """
    An App instance for managing Jobs. Jobs are containers for async processes
    such as data import or training.
    """
    def __init__(self, app):
        self.app = app

    def create_clip(self, asset, timeline, track, start, stop, content):
        """
        Create a new clip.  If a clip with the same metadata already exists it will
        simply be replaced.

        Args:
            asset (Asset): The asset or its unique Id.
            timeline (str): The timeline name for the clip.
            track (str): The track name for the clip.
            start (float): The starting point for the clip in seconds.
            stop (float): The ending point for the clip in seconds.
            content (str): The content of the clip.

        Returns:
            Clip: The clip that was created.
        """
        body = {
            "assetId": as_id(asset),
            "timeline": timeline,
            "track": track,
            "start": start,
            "stop": stop,
            "content": content
        }
        return VideoClip(self.app.client.post('/api/v1/clips', body))

    def create_clips(self, timeline):
        """
        Batch create clips using a TimelineBuilder.

        Args:
            timeline: (TimelineBuilder): A timeline builder.

        Returns:
            dict: A status dictionary
        """
        return self.app.client.post('/api/v1/clips/_timeline', timeline)

    def get_webvtt(self,
                   asset,
                   dst_file=None,
                   timeline=None,
                   track=None,
                   content=None):
        """
        Get all clip data as a WebVTT file and filter by specified options.

        Args:
            asset (Asset): The asset or unique Id.
            timeline: (str): A timeline name or collection of timeline names.
            track: (str): A track name or collection of track names.
            content (str): A content string to match.
            dst_file (mixed): An optional writable file handle or path to file.

        Returns:
            mixed: The text of the webvtt or the size of the written file.
        """
        body = {
            'assetId': as_id(asset),
            'timelines': as_id_collection(timeline),
            'tracks': as_id_collection(track),
            'content': as_id_collection(content)
        }
        rsp = self.app.client.post('/api/v1/clips/_webvtt', body=body, is_json=False)
        return self.__handle_webvtt(rsp, dst_file)

    def scroll_search(self, search=None, timeout="1m"):
        """
        Perform a VideoClip scrolled search using the ElasticSearch query DSL.

        See Also:
            For search/query format.
            https://www.elastic.co/guide/en/elasticsearch/reference/6.4/search-request-body.html

        Args:
            search (dict): The ElasticSearch search to execute
            timeout (str): The scroll timeout.  Defaults to 1 minute.
        Returns:
            VideoClipSearchScroller - an VideoClipSearchScroller instance which can be used as
            a generator for paging results.

        """
        return VideoClipSearchScroller(self.app, search, timeout)

    def search(self, search=None):
        """
        Perform an VideoClip search using the ElasticSearch query DSL.

        See Also:
            For search/query format.
            https://www.elastic.co/guide/en/elasticsearch/reference/6.4/search-request-body.html

        Args:
            search (dict): The ElasticSearch search to execute.
        Returns:
            VideoClipSearchResult - A VideoClipSearchResult instance.
        """
        return VideoClipSearchResult(self.app, search)

    def get_clip(self, id):
        """
        Get a VideoClip by unique Id.

        Args:
            id (str): The VideoClip or its unique Id.

        Returns:
            VideoClip: The clip with the given Id.
        """
        return VideoClip(self.app.client.get(f'api/v1/clips/{id}'))

    def __handle_webvtt(self, rsp, dst_file):
        """
        Handle a webvtt file response.

        Args:
            rsp (Response): A response from requests.
            dst_file (mixed): An optional file path or file handle.

        Returns:
            (mixed): Return the content itself or the content size if written to file.
        """
        if dst_file:
            if isinstance(dst_file, str):
                with open(dst_file, 'w') as fp:
                    fp.write(rsp.content.decode())
                return len(rsp.content)
            else:
                dst_file.write(rsp.content.decode())
                return len(rsp.content)
        else:
            return rsp.content.decode()



"""
Classes and functions for building timelines.
"""
from ..entity.asset import StoredFile
from ..util import as_id, as_collection

__all__ = [
    'TimelineBuilder',
    'VideoClip'
]


class VideoClip:
    """
    Clips represent a prediction for a section of video.
    """
    def __init__(self, data):
        self._data = data

    @property
    def id(self):
        """The Asset id the clip is associated with."""
        return self._data['id']

    @property
    def asset_id(self):
        """The Asset id the clip is associated with."""
        return self._data['assetId']

    @property
    def timeline(self):
        """The name of the timeline, this is the same as the pipeline module."""
        return self._data['timeline']

    @property
    def track(self):
        """The track name"""
        return self._data['track']

    @property
    def content(self):
        """The content of the clip. This is the prediction"""
        return self._data['content']

    @property
    def length(self):
        """The length of the clip"""
        return self._data['length']

    @property
    def start(self):
        """The start time of the clip"""
        return self._data['start']

    @property
    def stop(self):
        """The stop time of the clip"""
        return self._data['stop']

    @property
    def score(self):
        """The prediction score"""
        return self._data['score']

    @property
    def simhash(self):
        """A similarity hash, if any"""
        return self._data.get('simhash')

    @property
    def files(self):
        """The array of associated files."""
        return [StoredFile(f) for f in self._data.get('files', [])]

    @staticmethod
    def from_hit(hit):
        """
        Converts an ElasticSearch hit into an VideoClip.

        Args:
            hit (dict): An raw ES document

        Returns:
            Asset: The Clip.
        """
        data = {
            'id': hit['_id'],
        }
        data.update(hit.get('_source', {}).get('clip', {}))
        return VideoClip(data)

    def __len__(self):
        return self.length

    def __str__(self):
        return "<VideoClip id='{}'/>".format(self.id)

    def __repr__(self):
        return "<VideoClip id='{}' at {}/>".format(self.id, hex(id(self)))

    def __eq__(self, other):
        return other.id == self.id

    def __hash__(self):
        return hash(self.id)


class TimelineBuilder:
    """
    The TimelineBuilder class is used for batch creation of video clips.  Clips within a track
    can be overlapping.  Duplicate clips are automatically compacted to the highest score.
    """

    def __init__(self, asset, name):
        """
        Create a new timeline instance.
        Args:
            name (str): The name of the Timeline.
        """
        self.asset = as_id(asset)
        self.name = name
        self.tracks = {}

    def add_clip(self, track_name, start, stop, content, score=1, tags=None):
        """
        Add a clip to the timeline.

        Args:
            track_name (str): The Track name.
            start (float): The starting time.
            stop (float): The end time.
            content (str): The content.
            score: (float): The score if any.
            tags: (list): A list of tags that describes the content.

        Returns:
            (dict): A clip entry.

        """
        if stop < start:
            raise ValueError("The stop time cannot be smaller than the start time.")

        track = self.tracks.get(track_name)
        if not track:
            track = {'name': track_name, 'clips': []}
            self.tracks[track_name] = track

        clip = {
            "start": start,
            "stop":  stop,
            "content": [c.replace("\n", " ").strip() for c in as_collection(content)],
            "score": score,
            "tags": as_collection(tags)
        }

        track['clips'].append(clip)
        return clip

    def for_json(self):
        return {
            'name': self.name,
            'assetId': self.asset,
            'tracks': [track for track in self.tracks.values() if track['clips']]
        }



import copy

from .entity import VideoClip, Asset, ZmlpException
from .util import as_collection

__all__ = [
    'AssetSearchScroller',
    'VideoClipSearchScroller',
    'AssetSearchResult',
    'VideoClipSearchResult',
    'AssetSearchCsvExporter',
    'LabelConfidenceQuery',
    'SingleLabelConfidenceQuery',
    'SimilarityQuery',
    'FaceSimilarityQuery',
    'LabelConfidenceTermsAggregation',
    'LabelConfidenceMetricsAggregation'
]


class SearchScroller:
    """
    The SearchScroller can iterate over large amounts of data without incurring paging
    overhead by utilizing a server side cursor.  The cursor is held open for the specified
    timeout time unless it is refreshed before the timeout occurs.  In this sense, it's important
    to complete whatever operation you're taking on each asset within the timeout time.  For example
    if your page size is 32 and your timeout is 1m, you have 1 minute to handles 32 assets.  If that
    is not enough time, consider increasing the timeout or lowering your page size.
    """
    def __init__(self, klass, endpoint, app, search, timeout="1m", raw_response=False):
        """
        Create a new AbstractSearchScroller instance.

        Args:
            app (ZmlpApp): A ZmlpApp instance.
            search: (dict): The ES search
            timeout (str): The maximum amount of time the ES scroll will be active unless it's
                refreshed.
            raw_response (bool): Yield the raw ES response rather than assets. The raw
                response will contain the entire page, not individual assets.
        """
        self.klass = klass
        self.endpoint = endpoint
        self.app = app
        if search and getattr(search, "to_dict", None):
            search = search.to_dict()
        self.search = copy.deepcopy(search or {})
        self.timeout = timeout
        self.raw_response = raw_response

    def batches_of(self, batch_size=50):
        """
        A generator function capable of efficiently scrolling through
        large numbers of assets, returning them in batches of
        the given batch size.

        Args:
            batch_size (int): The size of the batch.

        Returns:
            generator: A generator that yields batches of Assets.

        """
        batch = []
        for asset in self.scroll():
            batch.append(asset)
            if len(batch) >= batch_size:
                yield batch
                batch = []
        if batch:
            yield batch

    def scroll(self):
        """
        A generator function capable of efficiently scrolling through large
        results.

        Examples:
            for asset in AssetSearchScroller({"query": {"term": { "source.extension": "jpg"}}}):
                do_something(asset)

        Yields:
            Asset: Assets that matched the search
        """
        result = self.app.client.post(
            "{}?scroll={}".format(self.endpoint, self.timeout), self.search)
        scroll_id = result.get("_scroll_id")
        if not scroll_id:
            raise ZmlpException("No scroll ID returned with scroll search, has it timed out?")
        try:
            while True:
                hits = result.get("hits")
                if not hits:
                    return
                if self.raw_response:
                    yield result
                else:
                    for hit in hits['hits']:
                        yield self.klass.from_hit(hit)

                scroll_id = result.get("_scroll_id")
                if not scroll_id:
                    raise ZmlpException(
                        "No scroll ID returned with scroll search, has it timed out?")
                result = self.app.client.post("api/v3/assets/_search/scroll", {
                    "scroll": self.timeout,
                    "scroll_id": scroll_id
                })
                if not result["hits"]["hits"]:
                    return
        finally:
            self.app.client.delete("api/v3/assets/_search/scroll", {
                "scroll_id": scroll_id
            })

    def __iter__(self):
        return self.scroll()


class AssetSearchScroller(SearchScroller):
    """
    AssetSearchScroller handles scrolling through Assets.
    """
    def __init__(self, app, search, timeout="1m", raw_response=False):
        super(AssetSearchScroller, self).__init__(
            Asset, 'api/v3/assets/_search', app, search, timeout, raw_response
        )


class VideoClipSearchScroller(SearchScroller):
    """
    VideoClipSearchScroller handles scrolling through video clips.
    """

    def __init__(self, app, search, timeout="1m", raw_response=False):
        super(VideoClipSearchScroller, self).__init__(
            VideoClip, 'api/v1/clips/_search', app, search, timeout, raw_response
        )


class AssetSearchCsvExporter:
    """
    Export a search to a CVS file.
    """
    def __init__(self, app, search):
        self.app = app
        self.search = search

    def export(self, fields, path):
        """
        Export the given fields to a csv file output path.

        Args:
            fields (list): An array of field names.
            path (str): a file path.

        Returns:
            int: The number of assets exported.

        """
        count = 0
        scroller = AssetSearchScroller(self.app, self.search)
        fields = as_collection(fields)
        with open(str(path), "w") as fp:
            for asset in scroller:
                count += 1
                line = ",".join(["'{}'".format(asset.get_attr(field)) for field in fields])
                fp.write(f'{line}\n')
        return count


class SearchResult:
    """
    Stores a search result from ElasticSearch and provides some convenience methods
    for accessing the data.

    """
    def __init__(self, klass, endpoint, app, search):
        """
        Create a new SearchResult.

        Args:
            klass (Class): The Class to wrap the search result.
            endpoint (str): The endpoint to use for search.
            app (ZmlpApp): A ZmlpApp instance.
            search (dict): An ElasticSearch query.
        """
        self.klass = klass
        self.endpoint = endpoint
        self.app = app
        if search and getattr(search, "to_dict", None):
            search = search.to_dict()
        self.search = search
        self.result = None

        self._execute_search()

    @property
    def items(self):
        """
        A list of assets returned by the query. This is not all of the matches,
        just a single page of results.

        Returns:
            list: The list of assets for this page.

        """
        hits = self.result.get("hits")
        if not hits:
            return []
        return [self.klass.from_hit(hit) for hit in hits['hits']]

    def batches_of(self, batch_size, max_assets=None):
        """
        A generator function which returns batches of assets in the
        given batch size.  This method will optionally page through
        N pages, yielding arrays of assets as it goes.

        This method is preferred to scrolling for Assets when
        multiple pages of Assets need to be processed.

        Args:
            batch_size (int): The size of the batch.
            max_assets (int): The max number of assets to return, max is 10k

        Returns:
            generator: A generator that yields batches of Assets.

        """
        # The maximum we can page through is 10k
        asset_countdown = max_assets or 10000

        batch = []
        while True:
            assets = self.assets
            if not assets:
                break

            for asset in assets:
                batch.append(asset)
                asset_countdown -= 1
                if asset_countdown <= 0:
                    break
                if len(batch) >= batch_size:
                    yield batch
                    batch = []

            if asset_countdown <= 0:
                break

            self.search['from'] = self.search.get('from', 0) + len(assets)
            self._execute_search()

        if batch:
            yield batch

    def aggregation(self, name):
        """
        Return an aggregation dict with the given name.

        Args:
            name (str): The agg name

        Returns:
            dict: the agg dict or None if no agg exists.
        """
        aggs = self.result.get("aggregations")
        if not aggs:
            return None

        if "#" in name:
            key = [name]
        else:
            key = [k for k in
                   self.result.get("aggregations", {}) if k.endswith("#{}".format(name))]

        if len(key) > 1:
            raise ValueError(
                "Aggs with the same name must be qualified by type (pick 1):  {}".format(key))
        elif not key:
            return None
        try:
            return aggs[key[0]]
        except KeyError:
            return None

    def aggregations(self):
        """
        Return a dictionary of all aggregations.

        Returns:
            dict: A dict of aggregations keyed on name.
        """
        return self.result.get("aggregations", {})

    @property
    def size(self):
        """
        The number assets in this page.  See "total_size" for the total number of assets matched.

        Returns:
            int: The number of assets in this page.

        """
        return len(self.result["hits"]["hits"])

    @property
    def total_size(self):
        """
        The total number of assets matched by the query.

        Returns:
            long: The total number of assets matched.

        """
        return self.result["hits"]["total"]["value"]

    @property
    def raw_response(self):
        """
        The raw ES response.
        Returns:
            (dict) The raw SearchResponse returned by ElasticSearch

        """
        return self.result

    def next_page(self):
        """
        Return an AssetSearchResult containing the next page.

        Returns:
            AssetSearchResult: The next page

        """
        search = copy.deepcopy(self.search or {})
        search['from'] = search.get('from', 0) + len(self.result.get("hits"))
        return SearchResult(self.klass, self.endpoint, self.app, search)

    def _execute_search(self):
        self.result = self.app.client.post(self.endpoint, self.search)

    def __iter__(self):
        return iter(self.items)

    def __getitem__(self, item):
        return self.items[item]


class AssetSearchResult(SearchResult):
    """
    The AssetSearchResult subclass handles paging throug an Asset search result.
    """
    def __init__(self, app, search):
        super(AssetSearchResult, self).__init__(
            Asset, 'api/v3/assets/_search', app, search
        )

    @property
    def assets(self):
        return self.items


class VideoClipSearchResult(SearchResult):
    """
    The VideoClipSearchResult subclass handles paging through an VideoClip search result.
    """
    def __init__(self, app, search):
        super(VideoClipSearchResult, self).__init__(
            VideoClip, 'api/v1/clips/_search', app, search
        )

    @property
    def clips(self):
        return self.items


class LabelConfidenceTermsAggregation:
    """
    Convenience class for making a simple terms aggregation on an array of predictions
    """
    def __init__(self, namespace):
        self.field = "analysis.{}.predictions".format(namespace)

    def for_json(self):
        return {
            "nested": {
                "path": self.field
            },
            "aggs": {
                "names": {
                    "terms": {
                        "field": self.field + ".label",
                        "size": 1000,
                        "order": {"_count": "desc"}
                    }
                }
            }
        }


class LabelConfidenceMetricsAggregation(object):

    def __init__(self, namespace, agg_type="stats"):
        """
        Create a new LabelConfidenceMetricsAggregation

        Args:
            namespace (str): The analysis namespace. (ex: zvi-label-detection)
            agg_type (str): A type of metrics agg to perform.
                stats, extended_stats,
        """
        self.field = "analysis.{}.predictions".format(namespace)
        self.agg_type = agg_type

    def for_json(self):
        return {
            "nested": {
                "path": self.field
            },
            "aggs": {
                "labels": {
                    "terms": {
                        "field": self.field + ".label",
                        "size": 1000,
                        "order": {"_count": "desc"}
                    },
                    "aggs": {
                        "stats": {
                            self.agg_type: {
                                "field": self.field + ".score"
                            }
                        }
                    }
                }
            }
        }


class LabelConfidenceQuery(object):
    """
    A helper class for building a label confidence score query.  This query must point
    at label confidence structure:  For example: analysis.zvi.label-detection.

    References:
        "labels": [
                {"label": "dog", "score": 0.97 },
                {"label": "fox", "score": 0.63 }
        ]
    """

    def __init__(self, namespace, labels, min_score=0.1, max_score=1.0):
        """
        Create a new LabelConfidenceScoreQuery.

        Args:
            namespace (str): The analysis namespace with predictions. (ex: zvi-label-detection)
            labels (list): A list of labels to filter.
            min_score (float): The minimum label score, default to 0.1.
            Note that 0.0 allows everything.
            max_score (float): The maximum score, defaults to 1.0 which is highest
        """
        self.namespace = namespace
        self.field = "analysis.{}.predictions".format(namespace)
        self.labels = as_collection(labels)
        self.score = [min_score, max_score]

    def for_json(self):
        return {
            "bool": {
                "filter": [
                    {
                        "terms": {
                            self.field + ".label": self.labels
                        }
                    }
                ],
                "must": [
                    {
                        "nested": {
                            "path": self.field,
                            "query": {
                                "function_score": {
                                    "boost_mode": "sum",
                                    "field_value_factor": {
                                        "field": self.field + ".score",
                                        "missing": 0
                                    },
                                    "query": {
                                        "bool": {
                                            "filter": [
                                                {
                                                    "terms": {
                                                        self.field + ".label": self.labels
                                                    }
                                                },
                                                {
                                                    "range": {
                                                        self.field + ".score": {
                                                            "gte": self.score[0],
                                                            "lte": self.score[1]
                                                        }
                                                    }
                                                }
                                            ]
                                        }
                                    }
                                }
                            }
                        }
                    }
                ]
            }
        }


class SingleLabelConfidenceQuery(object):
    """
    A helper class for building a label confidence score query.  This query must point
    at label confidence structure:  For example: analysis.zvi.label-detection.

    References:
        "labels": [
                {"label": "dog", "score": 0.97 },
                {"label": "fox", "score": 0.63 }
        ]
    """

    def __init__(self, namespace, labels, min_score=0.1, max_score=1.0):
        """
        Create a new SingleLabelConfidenceScoreQuery.

        Args:
            namespace (str): The analysis namespace with predictions. (ex: zvi-label-detection)
            labels (list): A list of labels to filter.
            min_score (float): The minimum label score, default to 0.1.
            Note that 0.0 allows everything.
            max_score (float): The maximum score, defaults to 1.0 which is highest
        """
        self.namespace = namespace

        self.field = "analysis.{}".format(namespace)
        self.labels = as_collection(labels)
        self.score = [min_score, max_score]

    def for_json(self):
        return {
            "bool": {
                "filter": [
                    {
                        "terms": {
                            self.field + ".label": self.labels
                        }
                    }
                ],
                "must": [
                    {
                        "function_score": {
                            "query": {
                                "bool": {
                                    "must": [
                                        {
                                            "terms": {
                                                self.field + ".label": self.labels
                                            }
                                        },
                                        {
                                            "range": {
                                                self.field + ".score": {
                                                    "gte": self.score[0],
                                                    "lte": self.score[1]
                                                }
                                            }
                                        }
                                    ]
                                }
                            },
                            "boost": "5",
                            "boost_mode": "sum",
                            "field_value_factor": {
                                "field": self.field + ".score",
                                "missing": 0
                            }
                        }
                    }
                ]
            }
        }


class SimilarityQuery:
    """
    A helper class for building a similarity search.  You can embed this class anywhere
    in a ES query dict, for example:

    References:
        {
            "query": {
                "bool": {
                    "must": [
                        SimilarityQuery(hash_string)
                    ]
                }
            }
        }
    """
    def __init__(self, hashes, min_score=0.75, boost=1.0,
                 field="analysis.zvi-image-similarity.simhash"):
        self.field = field
        self.hashes = []
        self.min_score = min_score
        self.boost = boost

        self.add_hash(hashes)

    def add_hash(self, hashes):
        """
        Add a new hash to the search.

        Args:
            hashes (mixed): A similarity hash string or an asset.

        Returns:
            SimilarityQuery: this instance of SimilarityQuery
        """
        for simhash in as_collection(hashes) or []:
            if isinstance(simhash, Asset):
                self.hashes.append(simhash.get_attr(self.field))
            elif isinstance(simhash, VideoClip):
                if simhash.simhash:
                    self.hashes.append(simhash.simhash)
            else:
                self.hashes.append(simhash)
        return self

    def add_asset(self, asset):
        """
        See add_hash which handles both hashes and Assets.
        """
        return self.add_hash(asset)

    def for_json(self):
        return {
            "script_score": {
                "query": {
                    "match_all": {}
                },
                "script": {
                    "source": "similarity",
                    "lang": "zorroa-similarity",
                    "params": {
                        "minScore": self.min_score,
                        "field": self.field,
                        "hashes":  self.hashes
                    }
                },
                "boost": self.boost,
                "min_score": self.min_score
            }
        }

    def __add__(self, simhash):
        self.add_hash(simhash)
        return self


class VideoClipSimilarityQuery(SimilarityQuery):
    def __init__(self, hashes, min_score=0.75, boost=1.0):
        super(VideoClipSimilarityQuery, self).__init__(
            hashes, min_score, boost, 'clip.simhash')


class FaceSimilarityQuery:
    """
    Performs a face similarity search.
    """
    def __init__(self, faces, min_score=0.90, boost=1.0,
                 field="analysis.zvi-face-detection.predictions.simhash"):
        """
        Create a new FaceSimilarityQuery.

        Args:
            faces (list): A prediction with a 'simhash' property or a simhash itself.
            min_score (float): The minimum score.
            boost (float): A boost value which weights this query higer than others.
            field (str): An optional field to compare make the comparison with. Defaults to ZVI.
        """
        hashes = []
        for face in as_collection(faces):
            if isinstance(face, str):
                hashes.append(face)
            else:
                hashes.append(face['simhash'])

        self.simquery = SimilarityQuery(
            hashes,
            min_score,
            boost,
            field)

    def for_json(self):
        return self.simquery.for_json()



"""Classes to support training with Models"""
import os
import logging
import json

logger = logging.getLogger(__name__)

__all__ = [
    'TrainingSetDownloader'
]


class TrainingSetDownloader:
    """
    The TrainingSetDownloader class handles writing out the images labeled
    for model training to local disk.  The Assets are automatically sorted
    into train and validation sets.

    Multiple directory layouts are supported based on the Model type.

    Examples:

        # Label Detection Layout
        base_dir/flowers/set_validate/daisy
        base_dir/flowers/set_validate/rose
        base_dir/flowers/set_validate/daisy
        base_dir/flowers/set_validate/rose

        # Object Detection Layout is a COCO compatible layout
        base_dir/set_train/images/*
        base_dir/set_train/annotations.json
        base_dir/set_test/images/*
        base_dir/set_test/annotations.json
    """

    SET_TRAIN = "train"
    """Directory name for training images"""

    SET_VALIDATION = "validate"
    """Directory name for test images"""

    def __init__(self, app,  model, style, dst_dir, validation_split=0.2):
        """
        Create a new TrainingImageDownloader.

        Args:
            app: (ZmlpApp): A ZmlpApp instance.
            model: (Model): A Model or unique Model ID.
            style: (str): The output style: labels_std, objects_keras, objects_coco
            dst_dir (str): A destination directory to write the files into.
            validation_split (float): The number of images in the training
                set for every image in the validation set.
        """
        self.app = app
        self.model = app.models.get_model(model)
        self.style = style
        self.dst_dir = dst_dir
        self.validation_split = validation_split

        self.labels = {}
        self.label_distrib = {}

        self.query = {
            'size': 64,
            '_source': ['labels', 'files'],
            'query': {
                'nested': {
                    'path': 'labels',
                    'query': {
                        'bool': {
                            'must': [
                                {'term': {'labels.modelId': self.model.id}},
                                {'term': {'labels.scope': 'TRAIN'}}
                            ]
                        }
                    }
                }
            }
        }

        os.makedirs(self.dst_dir, exist_ok=True)

    def build(self, pool=None):
        """
        Downloads the files labeled for training a Model to local disk.

        Args:
            labels_std, objects_keras, objects_coco
            pool (multiprocessing.Pool): An optional Pool instance which can be used
                to download files in parallel.

        """
        if self.style == 'labels-standard':
            self._build_labels_std_format(pool)
        elif self.style == 'objects_coco':
            self._build_objects_coco_format(pool)
        elif self.style == 'objects_keras':
            self._build_objects_keras_format(pool)
        else:
            raise ValueError('{} not supported by the TrainingSetDownloader'.format(format))

    def _build_labels_std_format(self, pool):

        self._setup_labels_std_base_dir()

        for num, asset in enumerate(self.app.assets.scroll_search(self.query, timeout='5m')):
            prx = asset.get_thumbnail(0)
            if not prx:
                logger.warning('{} did not have a suitable thumbnail'.format(asset))
                continue

            ds_labels = self._get_labels(asset)
            if not ds_labels:
                logger.warning('{} did not have any labels'.format(asset))
                continue

            label = ds_labels[0].get('label')
            if not label:
                logger.warning('{} was not labeled.'.format(asset))
                continue

            dir_name = self._get_image_set_type(label)
            dst_path = os.path.join(self.dst_dir, dir_name, label, prx.cache_id)
            os.makedirs(os.path.dirname(dst_path), exist_ok=True)

            logger.info('Downloading to {}'.format(dst_path))
            if pool:
                pool.apply_async(self.app.assets.download_file, args=(prx, dst_path))
            else:
                self.app.assets.download_file(prx, dst_path)

    def _build_objects_coco_format(self, pool=None):
        """
        Write a labeled assets in a COCO object detection training structure.

        Args:
            pool (multiprocessing.Pool): A multi-processing pool for downloading really fast.

        Returns:
            str: A path to an annotation file.

        """

        self._setup_objects_coco_base_dir()

        coco = CocoAnnotationFileBuilder()

        for image_id, asset in enumerate(self.app.assets.scroll_search(self.query, timeout='5m')):
            prx = asset.get_thumbnail(1)
            if not prx:
                logger.warning('{} did not have a suitable thumbnail'.format(asset))
                continue

            ds_labels = self._get_labels(asset)
            if not ds_labels:
                logger.warning('{} did not have any labels'.format(asset))
                continue

            for label in ds_labels:

                set_type = self._get_image_set_type(label['label'])
                dst_path = os.path.join(self.dst_dir, set_type, 'images', prx.cache_id)
                if not os.path.exists(dst_path):
                    self._download_file(prx, dst_path, pool)

                image = {
                    'file_name': dst_path,
                    'height': prx.attrs['height'],
                    'width': prx.attrs['width']
                }

                category = {
                    'supercategory': 'none',
                    'name': label['label']
                }

                bbox, area = self._zvi_to_cocos_bbox(prx, label['bbox'])
                annotation = {
                    'bbox': bbox,
                    'segmentation': [],
                    'ignore': 0,
                    'area': area,
                    'iscrowd': 0
                }

                if set_type == self.SET_TRAIN:
                    coco.add_to_training_set(image, category, annotation)
                else:
                    coco.add_to_validation_set(image, category, annotation)

        # Write out the annotations files.
        with open(os.path.join(self.dst_dir, self.SET_TRAIN, "annotations.json"), "w") as fp:
            logger.debug("Writing training set annotations to {}".format(fp.name))
            json.dump(coco.get_training_annotations(), fp)

        with open(os.path.join(self.dst_dir, self.SET_VALIDATION, "annotations.json"), "w") as fp:
            logger.debug("Writing test set annotations to {}".format(fp.name))
            json.dump(coco.get_validation_annotations(), fp)

    def _build_objects_keras_format(self, pool=None):
        self._setup_objects_keras_base_dir()

        fp_train = open(os.path.join(self.dst_dir, self.SET_TRAIN, "annotations.csv"), "w")
        fp_test = open(os.path.join(self.dst_dir, self.SET_VALIDATION, "annotations.csv"), "w")
        unique_labels = set()

        try:
            search = self.app.assets.scroll_search(self.query, timeout='5m')
            for image_id, asset in enumerate(search):
                prx = asset.get_thumbnail(1)
                if not prx:
                    logger.warning('{} did not have a suitable thumbnail'.format(asset))
                    continue

                ds_labels = self._get_labels(asset)
                if not ds_labels:
                    logger.warning('{} did not have any labels'.format(asset))
                    continue

                for label in ds_labels:
                    unique_labels.add(label['label'])
                    set_type = self._get_image_set_type(label['label'])
                    dst_path = os.path.join(self.dst_dir, set_type, 'images', prx.cache_id)
                    if not os.path.exists(dst_path):
                        self._download_file(prx, dst_path, pool)
                    line = [
                        dst_path
                    ]
                    line.extend([str(point) for point in
                                 self._zvi_to_keras_bbox(prx, label['bbox'])])
                    line.append(label['label'])
                    str_line = "{}\n".format(",".join(line))
                    if set_type == self.SET_TRAIN:
                        fp_train.write(str_line)
                    else:
                        fp_test.write(str_line)
        finally:
            fp_train.close()
            fp_test.close()

        with open(os.path.join(self.dst_dir, "classes.csv"), "w") as fp_classes:
            for idx, cls in enumerate(sorted(unique_labels)):
                fp_classes.write("{},{}\n".format(cls, idx))

    def _zvi_to_keras_bbox(self, prx, bbox):
        total_width = prx.attrs['width']
        total_height = prx.attrs['height']
        return [int(total_width * bbox[0]),
                int(total_height * bbox[1]),
                int(total_width * bbox[2]),
                int(total_height * bbox[3])]

    def _zvi_to_cocos_bbox(self, prx, bbox):
        """
        Converts a ZVI bbox to a COCOs bbox.  The format is x, y, width, height.

        Args:
            prx (StoredFile): A StoredFile containing a proxy image.
            bbox (list): A ZVI bbox.

        Returns:
            list[float]: A COCOs style bbox.
        """
        total_width = prx.attrs['width']
        total_height = prx.attrs['height']
        pt = total_width * bbox[0], total_height * bbox[1]

        new_bbox = [
            int(pt[0]),
            int(pt[1]),
            int(abs(pt[0] - (total_width * bbox[2]))),
            int(abs(pt[0] - (total_height * bbox[3])))
        ]
        area = (new_bbox[2] - new_bbox[0]) * (new_bbox[3] - new_bbox[1])
        return new_bbox, area

    def _download_file(self, prx, dst_path, pool=None):
        if pool:
            pool.apply_async(self.app.assets.download_file, args=(prx, dst_path))
        else:
            self.app.assets.download_file(prx, dst_path)

    def _setup_labels_std_base_dir(self):
        """
        Sets up a directory structure for storing files used to train a model..

        The structure is basically:
            train/<label>/<img file>
            validate/<label>/<img file>
        """
        self.labels = self.app.models.get_label_counts(self.model)

        # This is layout #1, we need to add darknet layout for object detection.
        dirs = (self.SET_TRAIN, self.SET_VALIDATION)
        for set_name in dirs:
            os.makedirs('{}/{}'.format(self.dst_dir, set_name), exist_ok=True)
            for label in self.labels.keys():
                os.makedirs(os.path.join(self.dst_dir, set_name, label), exist_ok=True)

        logger.info('TrainingSetDownloader setup, using {} labels'.format(len(self.labels)))

    def _setup_objects_coco_base_dir(self):
        dirs = (self.SET_TRAIN, self.SET_VALIDATION)
        for set_name in dirs:
            os.makedirs(os.path.join(self.dst_dir, set_name, 'images'), exist_ok=True)

    def _setup_objects_keras_base_dir(self):
        dirs = (self.SET_TRAIN, self.SET_VALIDATION)
        for set_name in dirs:
            os.makedirs(os.path.join(self.dst_dir, set_name, 'images'), exist_ok=True)

    def _get_image_set_type(self, label):
        """
        Using the validation_split property, determine if the current label
        would be in the training set or validation set.

        Args:
            label (str): The label name.

        Returns:
            str: Either 'validate' or 'train', depending on the validation_split property.

        """
        # Everything is in the training set.
        if self.validation_split <= 0.0:
            return self.SET_TRAIN

        ratio = int(1.0 / self.validation_split)
        value = self.label_distrib.get(label, 0) + 1
        self.label_distrib[label] = value

        if value % ratio == 0:
            return self.SET_VALIDATION
        else:
            return self.SET_TRAIN

    def _get_labels(self, asset):
        """
        Get the current model label for the given asset.

        Args:
            asset (Asset): The asset to check.

        Returns:
            list[dict]: The labels for training a model.

        """
        ds_labels = asset.get_attr('labels')
        if not ds_labels:
            return []
        result = []
        for ds_label in ds_labels:
            if ds_label.get('modelId') == self.model.id:
                result.append(ds_label)
        return result


class CocoAnnotationFileBuilder:
    """
    CocoAnnotationFileBuilder manages building a COCO annotations file for both
    a training set and test set.
    """

    def __init__(self):
        self.train_set = {
            "output": {
                "type": "instances",
                "images": [],
                "annotations": [],
                "categories": []
            },
            "img_set": {},
            "cat_set": {}
        }

        self.validation_set = {
            "output": {
                "type": "instances",
                "images": [],
                "annotations": [],
                "categories": []
            },
            "img_set": {},
            "cat_set": {}
        }

    def add_to_training_set(self, img, cat, annotation):
        """
        Add the image, category and annotation to the training set.

        Args:
            img (dict): A COCO image dict.
            cat (dict): A COCO category dict.
            annotation: (dict): A COCO annotation dict.
        """
        self._add_to_set(self.train_set, img, cat, annotation)

    def add_to_validation_set(self, img, cat, annotation):
        """
        Add the image, category and annotation to the test set.

        Args:
            img (dict): A COCO image dict.
            cat (dict): A COCO category dict.
            annotation: (dict): A COCO annotation dict.

        """
        self._add_to_set(self.validation_set, img, cat, annotation)

    def _add_to_set(self, dataset, img, cat, annotation):
        """
        Add the image, category and annotation to the given set.

        Args:
            dataset (dict): The set we're building.
            img (dict): A COCO image dict.
            cat (dict): A COCO category dict.
            annotation: (dict): A COCO annotation dict.
        """
        img_idmap = dataset['img_set']
        cat_idmap = dataset['cat_set']
        output = dataset['output']
        annots = output['annotations']

        img['id'] = img_idmap.get(img['file_name'], len(img_idmap))
        cat['id'] = cat_idmap.get(cat['name'], len(cat_idmap))
        annotation['id'] = len(annots)
        annotation['category_id'] = cat['id']
        annotation['image_id'] = img['id']

        if img['file_name'] not in img_idmap:
            img_idmap[img['file_name']] = img['id']
            output['images'].append(img)

        if cat['name'] not in cat_idmap:
            cat_idmap[cat['name']] = cat['id']
            output['categories'].append(cat)

        output['annotations'].append(annotation)

    def get_training_annotations(self):
        """
        Return a structure suitable for a COCO annotations file.

        Returns:
            dict: The training annoations.=
        """
        return self.train_set['output']

    def get_validation_annotations(self):
        """
        Return a structure suitable for a COCO annotations file.

        Returns:
            dict: The test annoations.
        """
        return self.validation_set['output']




from ..entity import Project


class ProjectApp(object):

    def __init__(self, app):
        self.app = app

    def get_project(self):
        """
        Return the current API Key's assigned project.

        Returns:
            Project
        """
        return Project(self.app.client.get("/api/v1/project"))



import functools
import re
import uuid
import decimal
import zipfile
import os


def is_valid_uuid(val):
    """
    Return true if the given value is a valid UUID.

    Args:
        val (str): a string which might be a UUID.

    Returns:
        bool: True if UUID

    """
    try:
        uuid.UUID(str(val))
        return True
    except ValueError:
        return False


def as_collection(value):
    """If the given value is not a collection of some type, return
    the value wrapped in a list.

    Args:
        value (:obj:`mixed`):

    Returns:
        :obj:`list` of :obj:`mixed`: The value wrapped in alist.

    """
    if value is None:
        return None
    if isinstance(value, (set, list, tuple)):
        return value
    return [value]


class ObjectView:
    """
    Wraps a dictionary and provides an object based view.

    """
    snake = re.compile(r'(?<!^)(?=[A-Z])')

    def __init__(self, d):
        d = dict([(self.snake.sub('_', k).lower(), v) for k, v in d.items()])
        self.__dict__ = d


def as_id(value):
    """
    If 'value' is an object, return the 'id' property, otherwise return
    the value.  This is useful for when you need an entity's unique Id
    but the user passed in an instance of the entity.

    Args:
        value (mixed): A string o an object with an 'id' property.

    Returns:
        str: The id property.
    """
    return getattr(value, 'id', value)


def as_id_collection(value):
    """If the given value is not a collection of some type, return
    the value wrapped in a list.  Additionally entity instances
    are resolved into their unique id.

    Args:
        value (:obj:`mixed`):

    Returns:
        list: A list of entity unique ids.

    """
    if value is None:
        return None
    if isinstance(value, (set, list, tuple, dict)):
        return [getattr(it, "id", it) for it in value]
    return [getattr(value, "id", value)]


def memoize(func):
    """
    Cache the result of the given function.

    Args:
        func (function): A function to wrap.

    Returns:
        function: a wrapped function
    """
    cache = func.cache = {}

    @functools.wraps(func)
    def memoized_func(*args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cache:
            cache[key] = func(*args, **kwargs)
        return cache[key]

    return memoized_func


def truncate(number, places):
    """
    Truncate a float to the given number of places.

    Args:
        number (float): The number to truncate.
        places (int): The number of plaes to preserve.

    Returns:
        Decimal: The truncated decimal value.
    """
    if not isinstance(places, int):
        raise ValueError('Decimal places must be an integer.')
    if places < 1:
        raise ValueError('Decimal places must be at least 1.')

    with decimal.localcontext() as context:
        context.rounding = decimal.ROUND_DOWN
        exponent = decimal.Decimal(str(10 ** - places))
        return decimal.Decimal(str(number)).quantize(exponent)


def round_all(items, precision=3):
    """
    Round all items in the list.

    Args:
        items (list): A list of floats.
        precision: (int): number of decimal places.

    Returns:
        list: A rounded list.
    """
    return [round(i, precision) for i in items]


def zip_directory(src_dir, dst_file, zip_root_name=""):
    """
    A utility function for ziping a directory of files.

    Args:
        src_dir (str): The source directory.
        dst_file (str): The destination file.s
        zip_root_name (str): A optional root directory to place files in the zip.
    Returns:
        str: The dst file.

    """

    def zipdir(path, ziph, root_name):
        for root, dirs, files in os.walk(path):
            for file in files:
                if file == ".DS_Store":
                    continue
                zip_entry = os.path.join(root_name, root.replace(path, ""), file)
                ziph.write(os.path.join(root, file), zip_entry)

    src_dir = os.path.abspath(src_dir)
    zipf = zipfile.ZipFile(dst_file, 'w', zipfile.ZIP_DEFLATED)
    zipdir(src_dir + '/', zipf, zip_root_name)
    zipf.close()
    return dst_file



import base64
import binascii
import datetime
import decimal
import json
import logging
import os
import random
import sys
import time
from io import IOBase
from urllib.parse import urljoin

import jwt
import requests

from .entity.exception import ZmlpException

logger = logging.getLogger(__name__)

DEFAULT_SERVER = 'https://api.zvi.zorroa.com'


class ZmlpClient(object):
    """
    ZmlpClient is used to communicate to a ZMLP API server.
    """

    def __init__(self, apikey, server, **kwargs):
        """
        Create a new ZmlpClient instance.

        Args:
            apikey: An API key in any supported form. (dict, base64 string, or open file handle)
            server: The url of the server to connect to. Defaults to https://api.zmlp.zorroa.com
            project_id: An optional project UUID for API keys with access to multiple projects.
            max_retries: Maximum number of retries to make if the API server
                is down, 0 for unlimited.
        """
        self.apikey = self.__load_apikey(apikey)
        self.server = server
        self.project_id = kwargs.get('project_id', os.environ.get("ZMLP_PROJECT"))
        self.max_retries = kwargs.get('max_retries', 3)
        self.verify = True

    def stream(self, url, dst):
        """
        Stream the given URL path to local dst file path.

        Args:
            url (str): The URL to stream
            dst (str): The destination file path
        """
        try:
            with open(dst, 'wb') as handle:
                response = requests.get(self.get_url(url), verify=self.verify,
                                        headers=self.headers(), stream=True)

                if not response.ok:
                    raise ZmlpClientException(
                        "Failed to stream asset: %s, %s" % (url, response))

                for block in response.iter_content(1024):
                    handle.write(block)
            return dst
        except requests.exceptions.ConnectionError as e:
            raise ZmlpConnectionException(e)

    def stream_text(self, url):
        """
        Stream the given URL.

        Args:
            url (str): The URL to stream

        Yields:
            generator (str): A generator of the lines making up the textual
                URL.
        """
        try:
            response = requests.get(self.get_url(url), verify=self.verify,
                                    headers=self.headers(), stream=True)
            if not response.ok:
                raise ZmlpClientException(
                    "Failed to stream text: %s" % response)

            for line in response.iter_lines(decode_unicode=True):
                yield line

        except requests.exceptions.ConnectionError as e:
            raise ZmlpConnectionException(e)

    def send_file(self, path, file_path):
        """
        Sends a file via request body

        Args:
            path (path): The URI fragment for the request.
            file_path (str): The path to the file to send.

        Returns:
            dict: A dictionary which can be used to fetch the file.
        """
        with open(file_path, 'rb') as f:
            return self.__handle_rsp(requests.post(
                self.get_url(path), headers=self.headers(content_type=""),
                data=f), True)

    def upload_file(self, path, file, body={}, json_rsp=True):
        """
        Upload a single file and a request to the given endpoint path.

        Args:
            path (str): The URL to upload to.
            file (str): The file path to upload.
            body (dict): A request body
            json_rsp (bool): Set to true if the result returned is JSON

        Returns:
            dict: The response body of the request.
        """
        try:
            post_files = [("file", (os.path.basename(file), open(file, 'rb')))]
            if body is not None:
                post_files.append(
                    ["body", (None, to_json(body), 'application/json')])

            return self.__handle_rsp(requests.post(
                self.get_url(path), headers=self.headers(content_type=""),
                files=post_files), json_rsp)

        except requests.exceptions.ConnectionError as e:
            raise ZmlpConnectionException(e)

    def upload_files(self, path, files, body, json_rsp=True):
        """
        Upload an array of files and a reques to the given endpoint path.

        Args:
            path (str): The URL to upload to
            files (list of str): The file paths to upload
            body (dict): A request body
            json_rsp (bool): Set to true if the result returned is JSON

        Returns:
            dict: The response body of the request.
        """
        try:
            post_files = []
            for f in files:
                if isinstance(f, IOBase):
                    post_files.append(
                        ("files", (os.path.basename(f.name), f)))
                else:
                    post_files.append(
                        ("files", (os.path.basename(f), open(f, 'rb'))))

            if body is not None:
                post_files.append(
                    ("body", ("", to_json(body),
                              'application/json')))

            return self.__handle_rsp(requests.post(
                self.get_url(path), headers=self.headers(content_type=""),
                verify=self.verify, files=post_files), json_rsp)

        except requests.exceptions.ConnectionError as e:
            raise ZmlpConnectionException(e)

    def get(self, path, body=None, is_json=True):
        """
        Performs a get request.

        Args:
            path (str): An archivist URI path.
            body (dict): The request body which will be serialized to json.
            is_json (bool): Set to true to specify a JSON return value

        Returns:
            object: The http response object or an object deserialized from the
                response json if the ``json`` argument is true.

        Raises:
            Exception: An error occurred making the request or parsing the
                JSON response
        """
        return self._make_request('get', path, body, is_json)

    def post(self, path, body=None, is_json=True):
        """
        Performs a post request.

        Args:
            path (str): An archivist URI path.
            body (object): The request body which will be serialized to json.
            is_json (bool): Set to true to specify a JSON return value

        Returns:
            object: The http response object or an object deserialized from the
                response json if the ``json`` argument is true.

        Raises:
            Exception: An error occurred making the request or parsing the
                JSON response
        """
        return self._make_request('post', path, body, is_json)

    def put(self, path, body=None, is_json=True):
        """
        Performs a put request.

        Args:
            path (str): An archivist URI path.
            body (object): The request body which will be serialized to json.
            is_json (bool): Set to true to specify a JSON return value

        Returns:
            object: The http response object or an object deserialized from the
                response json if the ``json`` argument is true.

        Raises:
            Exception: An error occurred making the request or parsing the
                JSON response
        """
        return self._make_request('put', path, body, is_json)

    def delete(self, path, body=None, is_json=True):
        """
         Performs a delete request.

         Args:
             path (str): An archivist URI path.
             body (object): The request body which will be serialized to json.
             is_json (bool): Set to true to specify a JSON return value

         Returns:
             object: The http response object or an object deserialized from
             the response json if the ``json`` argument is true.

         Raises:
             Exception: An error occurred making the request or parsing the
                JSON response
         """
        return self._make_request('delete', path, body, is_json)

    def iter_paged_results(self, url, req, limit, cls):
        """
        Handles paging through the results of the standard _search
        endpoints on the backend.

        Args:
            url (str): the URL to POST a search to
            req (object): the search request body
            limit (int): the maximum items to return, None for no limit.
            cls (type): the class to wrap each result in

        Yields:
            Generator

        """
        left_to_return = limit or sys.maxsize
        page = 0
        req["page"] = {}
        while True:
            if left_to_return < 1:
                break
            page += 1
            req["page"]["size"] = min(100, left_to_return)
            req["page"]["from"] = (page - 1) * req["page"]["size"]
            rsp = self.post(url, req)
            if not rsp.get("list"):
                break
            for f in rsp["list"]:
                yield cls(f)
                left_to_return -= 1
            # Used to break before pulling new batch
            if rsp.get("break"):
                break

    def _make_request(self, method, path, body=None, is_json=True):
        request_function = getattr(requests, method)
        if body is not None:
            data = to_json(body)
        else:
            data = body

        # Making the request is wrapped in its own try/catch so it's easier
        # to catch any and all socket and http exceptions that can possibly be
        # thrown.  Once hat happens, handle_rsp is called which may throw
        # application level exceptions.
        rsp = None
        tries = 0
        url = self.get_url(path, body)
        while True:
            try:
                rsp = request_function(url, data=data, headers=self.headers(),
                                       verify=self.verify)
                break
            except Exception as e:
                # Some form of connection error, wait until archivist comes
                # back.
                tries += 1
                if 0 < self.max_retries <= tries:
                    raise e
                wait = random.randint(1, random.randint(1, 60))
                # Switched to stderr in case no logger is setup, still want
                # to see messages.
                msg = "Communicating to ZMLP (%s) timed out %d times, " \
                      "waiting ... %d seconds, error=%s\n"
                sys.stderr.write(msg % (url, tries, wait, e))
                time.sleep(wait)

        return self.__handle_rsp(rsp, is_json)

    def __handle_rsp(self, rsp, is_json):
        if rsp.status_code != 200:
            self.__raise_exception(rsp)
        if is_json and len(rsp.content):
            rsp_val = rsp.json()
            if logger.getEffectiveLevel() == logging.DEBUG:
                logger.debug(
                    "rsp: status: %d  body: '%s'" % (rsp.status_code, rsp_val))
            return rsp_val
        return rsp

    def __raise_exception(self, rsp):
        data = {}
        try:
            data.update(rsp.json())
        except Exception as e:
            # The result is not json.
            data["message"] = "Your HTTP request was invalid '%s', response not " \
                              "JSON formatted. %s" % (rsp.status_code, e)
            data["status"] = rsp.status_code

        # If the status code can't be found, then ZmlpRequestException is returned.
        ex_class = translate_exception(rsp.status_code)
        raise ex_class(data)

    def get_url(self, path, body=None):
        """
        Returns the full URL including the configured server part.
        """
        url = urljoin(self.server, path)
        if logger.getEffectiveLevel() == logging.DEBUG:
            logger.debug("url: '%s' path: '%s' body: '%s'" % (url, path, body))
        return url

    def headers(self, content_type="application/json"):
        """
        Generate the return some request headers.

        Args:
            content_type(str):  The content-type for the request. Defaults to
                'application/json'

        Returns:
            dict: An http header struct.

        """
        header = {'Authorization': "Bearer {}".format(self.__sign_request())}

        if content_type:
            header['Content-Type'] = content_type

        if logger.getEffectiveLevel() == logging.DEBUG:
            logger.debug("headers: %s" % header)

        return header

    def __load_apikey(self, apikey):
        key_data = None
        if not apikey:
            return key_data
        elif hasattr(apikey, 'read'):
            key_data = json.load(apikey)
        elif isinstance(apikey, dict):
            key_data = apikey
        elif isinstance(apikey, (str, bytes)):
            try:
                key_data = json.loads(base64.b64decode(apikey))
            except binascii.Error:
                raise ValueError("Invalid base64 encoded API key.")

        return key_data

    def __sign_request(self):
        if not self.apikey:
            raise RuntimeError("Unable to make request, no ApiKey has been specified.")
        claims = {
            'aud': self.server,
            'exp': datetime.datetime.utcnow() + datetime.timedelta(seconds=60),
            'accessKey': self.apikey["accessKey"],
        }

        if os.environ.get("ZMLP_TASK_ID"):
            claims['taskId'] = os.environ.get("ZMLP_TASK_ID")
            claims['jobId'] = os.environ.get("ZMLP_JOB_ID")

        if self.project_id:
            claims["projectId"] = self.project_id
        return jwt.encode(claims, self.apikey['secretKey'], algorithm='HS512')


class SearchResult(object):
    """
    A utility class for wrapping various search result formats
    that come back from the ZMLP servers.
    """

    def __init__(self, data, clazz):
        """
        Create a new SearchResult instance.

        Note that its possible to both iterate and index a SearchResult
        as a list. For example

        Args:
            data (dict): A search response body from the ZMLP servers.
            clazz (mixed): A class to wrap each item in the response body.
        """
        self.items = [clazz(item) for item in data["list"]]
        self.offset = data["page"]["from"]
        self.size = len(data["list"])
        self.total = data["page"]["totalCount"]

    def __iter__(self):
        return iter(self.items)

    def __getitem__(self, idx):
        return self.items[idx]


def to_json(obj, indent=None):
    """
    Convert the given object to a JSON string using
    the ZmlpJsonEncoder.

    Args:
        obj (mixed): any json serializable python object.
        indent (int): The indentation level for the json, or None for compact.
    Returns:
        str: The serialized object

    """
    val = json.dumps(obj, cls=ZmlpJsonEncoder, indent=indent)
    if logger.getEffectiveLevel() == logging.DEBUG:
        logger.debug("json: %s" % val)
    return val


class ZmlpJsonEncoder(json.JSONEncoder):
    """
    JSON encoder for with ZMLP specific serialization defaults.
    """

    def default(self, obj):
        if hasattr(obj, 'for_json'):
            return obj.for_json()
        elif isinstance(obj, (set, frozenset)):
            return list(obj)
        elif isinstance(obj, datetime.datetime):
            return obj.isoformat()
        elif isinstance(obj, datetime.date):
            return obj.isoformat()
        elif isinstance(obj, datetime.time):
            return obj.isoformat()
        elif isinstance(obj, decimal.Decimal):
            return float(obj)

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


class ZmlpClientException(ZmlpException):
    """The base exception class for all ZmlpClient related Exceptions."""
    pass


class ZmlpRequestException(ZmlpClientException):
    """
    The base exception class for all exceptions thrown from zmlp.
    """
    def __init__(self, data):
        super(ZmlpClientException, self).__init__(
            data.get("message", "Unknown request exception"))
        self.__data = data

    @property
    def type(self):
        return self.__data["exception"]

    @property
    def cause(self):
        return self.__data["cause"]

    @property
    def endpoint(self):
        return self.__data["path"]

    @property
    def status(self):
        return self.__data["status"]

    def __str__(self):
        return "<ZmlpRequestException msg=%s>" % self.__data["message"]


class ZmlpConnectionException(ZmlpClientException):
    """
    This exception is thrown if the client encounters a connectivity issue
    with the Zmlp API servers..
    """
    pass


class ZmlpWriteException(ZmlpRequestException):
    """
    This exception is thrown the Zmlp fails a write operation.
    """

    def __init__(self, data):
        super(ZmlpWriteException, self).__init__(data)


class ZmlpSecurityException(ZmlpRequestException):
    """
    This exception is thrown if Zmlp fails a security check on the request.
    """

    def __init__(self, data):
        super(ZmlpSecurityException, self).__init__(data)


class ZmlpNotFoundException(ZmlpRequestException):
    """
    This exception is thrown if the Zmlp fails a read operation because
    a piece of named data cannot be found.
    """

    def __init__(self, data):
        super(ZmlpNotFoundException, self).__init__(data)


class ZmlpDuplicateException(ZmlpWriteException):
    """
    This exception is thrown if the Zmlp fails a write operation because
    the newly created element would be a duplicate.
    """

    def __init__(self, data):
        super(ZmlpDuplicateException, self).__init__(data)


class ZmlpInvalidRequestException(ZmlpRequestException):
    """
    This exception is thrown if the request sent to Zmlp is invalid in
    some way, similar to an IllegalArgumentException.
    """

    def __init__(self, data):
        super(ZmlpInvalidRequestException, self).__init__(data)


"""
A map of HTTP response codes to local exception types.
"""
EXCEPTION_MAP = {
    404: ZmlpNotFoundException,
    409: ZmlpDuplicateException,
    500: ZmlpInvalidRequestException,
    400: ZmlpInvalidRequestException,
    401: ZmlpSecurityException,
    403: ZmlpSecurityException
}


def translate_exception(status_code):
    """
    Translate the HTTP status code into one of the exceptions.

    Args:
        status_code (int): the HTTP status code

    Returns:
        Exception: the exception to throw for the given status code
    """
    return EXCEPTION_MAP.get(status_code, ZmlpRequestException)



import logging

from ..entity import AnalysisModule
from ..util import as_collection, as_id

logger = logging.getLogger(__name__)

__all__ = [
    'AnalysisModuleApp'
]


class AnalysisModuleApp:
    """
    App class for querying Analysis Modules
    """

    def __init__(self, app):
        self.app = app

    def get_analysis_module(self, id):
        """
        Get an AnalysisModule by Id.

        Args:
            id (str): The AnalysisModule ID or a AnalysisModule instance.

        Returns:
            AnalysisModule: The matching AnalysisModule
        """
        return AnalysisModule(self.app.client.get('/api/v1/pipeline-mods/{}'.format(as_id(id))))

    def find_one_analysis_module(self, id=None, name=None, type=None, category=None, provider=None):
        """
        Find a single AnalysisModule based on various properties.

        Args:
            id (str): The ID or list of Ids.
            name (str): The model name or list of names.
            type: (str): A AnalysisModule typ type or collection of types to filter on.
            category (str): The category of AnalysisModuleule
            provider (str): The provider of the AnalysisModuleule
        Returns:
            AnalysisModule: The matching AnalysisModule.
        """
        body = {
            'names': as_collection(name),
            'ids': as_collection(id),
            'types': as_collection(type),
            'categories': as_collection(category),
            'providers': as_collection(provider)
        }
        return AnalysisModule(self.app.client.post('/api/v1/pipeline-mods/_find_one', body))

    def find_analysis_modules(self, keywords=None, id=None, name=None, type=None,
                              category=None, provider=None, limit=None, sort=None):
        """
        Search for AnalysisModule.

        Args:
            keywords(str): Keywords that match various fields on a AnalysisModule
            id (str): An ID or collection of IDs to filter on.
            name (str): A name or collection of names to filter on.
            type: (str): A AnalysisModule type type or collection of types to filter on.
            category (str): The category or collection of category names.
            provider (str): The provider or collection provider names.
            limit: (int) Limit the number of results.
            sort: (list): A sort array, example: ["time_created:desc"]

        Returns:
            generator: A generator which will return matching AnalysisModules when iterated.
        """
        body = {
            'keywords': str(keywords),
            'names': as_collection(name),
            'ids': as_collection(id),
            'types': as_collection(type),
            'categories': as_collection(category),
            'providers': as_collection(provider),
            'sort': sort
        }
        return self.app.client.iter_paged_results(
            '/api/v1/pipeline-mods/_search', body, limit, AnalysisModule)



from datetime import datetime


class BaseEntity:

    def __init__(self, data):
        self._data = data

    @property
    def id(self):
        """The id of the Entity"""
        return self._data['id']

    @property
    def time_created(self):
        """The date/time the entity was created."""
        return datetime.fromtimestamp(self._data['timeCreated'] / 1000.0)

    @property
    def time_modified(self):
        """The date/time the entity was modified."""
        return datetime.fromtimestamp(self._data['timeModified'] / 1000.0)

    @property
    def actor_created(self):
        """The UUID of the actor that created the entity."""
        return self._data['actorCreated']

    @property
    def actor_modified(self):
        """The UUID of the actor that modified the entity."""
        return self._data['actorModified']

    def __hash__(self):
        return hash(self._data['id'])

    def __eq__(self, other):
        return self._data['id'] == getattr(other, 'id', None)

    def __str__(self):
        vals = [self.__class__.__name__, self.id]
        name = self._data.get('name')
        if name:
            vals.append(name)
            return "<{} id={} name={}>".format(*vals)
        else:
            return "<{} id={}>".format(*vals)



import logging
import os
import tempfile

from ..entity import Model, Job, ModelTypeInfo, AnalysisModule
from ..training import TrainingSetDownloader
from ..util import as_collection, as_id, zip_directory

logger = logging.getLogger(__name__)

__all__ = [
    'ModelApp'
]


class ModelApp:
    """
    Methods for manipulating models.
    """

    def __init__(self, app):
        self.app = app

    def create_model(self, name, type):
        """
        Create and retrn a new model .

        Args:
            name (str): The name of the model.
            type (ModelType): The type of Model, see the ModelType class.

        Returns:
            Model: The new model.
        """
        body = {
            "name": name,
            "type": type.name
        }
        return Model(self.app.client.post("/api/v3/models", body))

    def get_model(self, id):
        """
        Get a Model by Id
        Args:
            id (str): The model id.

        Returns:
            Model: The model.
        """
        return Model(self.app.client.get("/api/v3/models/{}".format(as_id(id))))

    def find_one_model(self, id=None, name=None, type=None):
        """
        Find a single Model based on various properties.

        Args:
            id (str): The ID or list of Ids.
            name (str): The model name or list of names.
            type (str): The model type or list of types.
        Returns:
            Model: the matching Model.
        """
        body = {
            'names': as_collection(name),
            'ids': as_collection(id),
            'types': as_collection(type)
        }
        return Model(self.app.client.post("/api/v3/models/_find_one", body))

    def find_models(self, id=None, name=None, type=None, limit=None, sort=None):
        """
        Find a single Model based on various properties.

        Args:
            id (str): The ID or list of Ids.
            name (str): The model name or list of names.
            type (str): The model type or list of types.
            limit (int): Limit results to the given size.
            sort (list): An arary of properties to sort by. Example: ["name:asc"]

        Returns:
            generator: A generator which will return matching Models when iterated.

        """
        body = {
            'names': as_collection(name),
            'ids': as_collection(id),
            'types': as_collection(type),
            'sort': sort
        }
        return self.app.client.iter_paged_results('/api/v3/models/_search', body, limit, Model)

    def train_model(self, model, deploy=False, **kwargs):
        """
        Train the given Model by kicking off a model training job.

        Args:
            model (Model): The Model instance or a unique Model id.
            deploy (bool): Deploy the model on your production data immediately after training.
            **kwargs (kwargs): Model training arguments which differ based on the model..

        Returns:
            Job: A model training job.
        """
        model_id = as_id(model)
        body = {
            'deploy': deploy,
            'args': dict(kwargs)
        }
        return Job(self.app.client.post('/api/v3/models/{}/_train'.format(model_id), body))

    def deploy_model(self, model, search=None, file_types=None):
        """
        Apply the model to the given search.

        Args:
            model (Model): A Model instance or a model unique Id.
            search (dict): An arbitrary asset search, defaults to using the
                deployment search associated with the model
            file_types (list): An optional file type filer, can be combination of
                "images", "documents", and "videos"

        Returns:
            Job: The Job that is hosting the reprocess task.
        """
        mid = as_id(model)
        body = {
            "search": search,
            "fileTypes": file_types,
            "jobId": os.environ.get("ZMLP_JOB_ID")
        }
        return Job(self.app.client.post(f'/api/v3/models/{mid}/_deploy', body))

    def upload_trained_model(self, model, model_path, labels):
        """
        Uploads a Tensorflow2/Keras model.  For the 'model_path' arg you can either
        pass the path to a Tensorflow saved model or a trained model instance itself.

        Args:
            model (Model): The Model or te unique Model ID.
            model_path (mixed): The path to the model directory or a Tensorflow model instance.
            labels (list): The list of labels,.
        Returns:
            AnalysisModule: The AnalysisModule configured to use the model.
        """

        if not labels:
            raise ValueError("Uploading a model requires an array of labels")

        # check to see if its a keras model and save to a temp dir.
        if getattr(model_path, 'save', None):
            tmp_path = tempfile.mkdtemp()
            model_path.save(tmp_path)
            model_path = tmp_path

        with open(model_path + '/labels.txt', 'w') as fp:
            for label in labels:
                fp.write(f'{label}\n')

        model_file = tempfile.mkstemp(prefix="model_", suffix=".zip")[1]
        zip_file_path = zip_directory(model_path, model_file)

        mid = as_id(model)
        return AnalysisModule(self.app.client.send_file(
            f'/api/v3/models/{mid}/_upload', zip_file_path))

    def get_label_counts(self, model):
        """
        Get a dictionary of the labels and how many times they occur.

        Args:
            model (Model): The Model or its unique Id.

        Returns:
            dict: a dictionary of label name to occurrence count.

        """
        return self.app.client.get('/api/v3/models/{}/_label_counts'.format(as_id(model)))

    def rename_label(self, model, old_label, new_label):
        """
        Rename a the given label to a new label name.  The new label can already exist.

        Args:
            model (Model): The Model or its unique Id.
            old_label (str): The old label name.
            new_label (str): The new label name.

        Returns:
            dict: a dictionary containing the number of assets updated.

        """
        body = {
            "label": old_label,
            "newLabel": new_label
        }
        return self.app.client.put('/api/v3/models/{}/labels'.format(as_id(model)), body)

    def delete_label(self, model, label):
        """
        Removes the label from all Assets.

        Args:
            model (Model): The Model or its unique Id.
            label (str): The label name to remove.

        Returns:
            dict: a dictionary containing the number of assets updated.

        """
        body = {
            "label": label
        }
        return self.app.client.delete('/api/v3/models/{}/labels'.format(as_id(model)), body)

    def download_labeled_images(self, model, style, dst_dir, validation_split=0.2):
        """
        Get a TrainingSetDownloader instance which can be used to download all the
        labeled images for a Model to local disk.

        Args:
            model (Model): The Model or its unique ID.
            style (str): The structure style to build: labels_std, objects_keras, objects_coco
            dst_dir (str): The destination dir to write the Assets into.
            validation_split (float): The ratio of training images to validation images.
                Defaults to 0.2.
        """
        return TrainingSetDownloader(self.app, model, style, dst_dir, validation_split)

    def get_model_type_info(self, model_type):
        """
        Get additional properties concerning a specific model type.

        Args:
            model_type (ModelType): The model type Enum or name.

        Returns:
            ModelTypeInfo: Additional properties related to a model type.
        """
        type_name = getattr(model_type, 'name', str(model_type))
        return ModelTypeInfo(self.app.client.get(f'/api/v3/models/_types/{type_name}'))

    def get_all_model_type_info(self):
        """
        Get all available ModelTypeInfo options.

        Returns:
            list: A list of ModelTypeInfo
        """
        return [ModelTypeInfo(info) for info in self.app.client.get('/api/v3/models/_types')]



from datetime import datetime

from .base import BaseEntity
from ..util import ObjectView

__all__ = [
    'Job',
    'Task',
    'TaskError'
]


class Job(BaseEntity):
    """
    A Job represents a backend data process.  Jobs are made up of Tasks
    which are scheduled to execute on Analyst data processing nodes.
    """

    def __init__(self, data):
        super(Job, self).__init__(data)

    @property
    def name(self):
        """The name of the Job"""
        return self._data['name']

    @property
    def state(self):
        """The state of the Job"""
        return self._data['state']

    @property
    def paused(self):
        """True if the Job is paused."""
        return self._data['paused']

    @property
    def priority(self):
        """The priority of the Job"""
        return self._data['priority']

    @property
    def time_started(self):
        """The datetime the job got the first analyst."""
        if self._data['timeStarted'] == -1:
            return None
        else:
            return datetime.fromtimestamp(self._data['timeStarted'] / 1000.0)

    @property
    def time_stopped(self):
        """The datetime the job finished."""
        if self._data['timeStopped'] == -1:
            return None
        else:
            return datetime.fromtimestamp(self._data['timeStopped'] / 1000.0)

    @property
    def asset_counts(self):
        """Asset counts for the Job"""
        return ObjectView(self._data['assetCounts'])

    @property
    def task_counts(self):
        """Task counts for the Job"""
        return ObjectView(self._data['taskCounts'])

    @property
    def time_modified(self):
        """The date/time the Job was modified."""
        return datetime.fromtimestamp(self._data['timeUpdated'] / 1000.0)


class Task(BaseEntity):
    """
    Jobs contain Tasks and each Task handles the processing for 1 or more files/assets.
    """

    def __init__(self, data):
        super(Task, self).__init__(data)

    @property
    def job_id(self):
        """The Job Id"""
        return self._data['jobId']

    @property
    def name(self):
        """The name of the Task"""
        return self._data['name']

    @property
    def state(self):
        """The name of the Task"""
        return self._data['state']

    @property
    def time_started(self):
        """The datetime the job got the first analyst."""
        if self._data['timeStarted'] == -1:
            return None
        else:
            return datetime.fromtimestamp(self._data['timeStarted'] / 1000.0)

    @property
    def time_stopped(self):
        """The datetime the job finished."""
        if self._data['timeStopped'] == -1:
            return None
        else:
            return datetime.fromtimestamp(self._data['timeStopped'] / 1000.0)

    @property
    def time_pinged(self):
        """The datetime the running task sent a watch dog ping."""
        if self._data['timePing'] == -1:
            return None
        else:
            return datetime.fromtimestamp(self._data['timePing'] / 1000.0)

    @property
    def time_modified(self):
        """The date/time the Job was modified."""
        return self.time_pinged

    @property
    def asset_counts(self):
        return ObjectView(self._data['assetCounts'])


class TaskError:
    """
    A TaskError contains information regarding a failed Task or Asset.
    """

    def __init__(self, data):
        self._data = data

    @property
    def id(self):
        """ID of the TaskError"""
        return self._data['id']

    @property
    def task_id(self):
        """UUID of the Task that encountered an error."""
        return self._data['taskId']

    @property
    def job_id(self):
        """UUID of the Job that encountered an error."""
        return self._data['jobId']

    @property
    def datasource_id(self):
        """UUID of the DataSource that encountered an error."""
        return self._data['dataSourceId']

    @property
    def asset_id(self):
        """ID of the Asset that encountered an error."""
        return self._data['assetId']

    @property
    def path(self):
        """File path or URI that was being processed."""
        return self._data['path']

    @property
    def message(self):
        """Error message from the exception that generated the error."""
        return self._data['message']

    @property
    def processor(self):
        """Processor in which the error occurred."""
        return self._data['processor']

    @property
    def fatal(self):
        """True if the error was fatal and the Asset was not processed."""
        return self._data['fatal']

    @property
    def phase(self):
        """Phase at which the error occurred: generate, execute, teardown."""
        return self._data['phase']

    @property
    def time_created(self):
        """The date/time the entity was created."""
        return datetime.fromtimestamp(self._data['timeCreated'] / 1000.0)

    @property
    def stack_trace(self):
        """Full stack trace from the error, if any."""
        return self._data['stackTrace']



from ..entity import DataSource, Job
from ..util import is_valid_uuid, as_collection


class DataSourceApp(object):

    def __init__(self, app):
        self.app = app

    def create_datasource(self, name, uri, modules=None, file_types=None, credentials=None):
        """
        Create a new DataSource.

        Args:
            name (str): The name of the data source.
            uri (str): The URI where the data can be found.
            modules (list): A list of AnalysisModules names to apply to the data.
            file_types (list of str): a list of file extensions or general types like
                'images', 'videos', 'documents'. Defaults to all file types.
            credentials (list of str): A list of pre-created credentials blob names.
        Returns:
            DataSource: The created DataSource

        """
        url = '/api/v1/data-sources'
        body = {
            'name': name,
            'uri': uri,
            'credentials': as_collection(credentials),
            'fileTypes': file_types,
            'modules': as_collection(modules)
        }
        return DataSource(self.app.client.post(url, body=body))

    def get_datasource(self, name):
        """
        Finds a DataSource by name or unique Id.

        Args:
            name (str): The unique name or unique ID.

        Returns:
            DataSource: The DataSource

        """
        url = '/api/v1/data-sources/_findOne'
        if is_valid_uuid(name):
            body = {"ids": [name]}
        else:
            body = {"names": [name]}

        return DataSource(self.app.client.post(url, body=body))

    def import_files(self, ds, batch_size=25):
        """
        Import all assets found at the given DataSource.  If the
        DataSource has already been imported then only new files will be
        imported. New modules assigned to the datasource will
        also be applied to existing assets as well as new assets.

        Args:
            ds (DataSource): A DataSource object or the name of a data source.
            batch_size (int): The number of Assets per batch.  Must be at least 20.
        Returns:
            Job: Return the Job responsible for processing the files.

        """
        body = {
            "batchSize": batch_size
        }
        url = '/api/v1/data-sources/{}/_import'.format(ds.id)
        return Job(self.app.client.post(url, body))

    def delete_datasource(self, ds, remove_assets=False):
        """
        Delete the given datasource.  If remove_assets is true, then all
        assets that were imported with a datasource are removed as well.  This
        cannot be undone.

        Args:
            ds (DataSource): A DataSource object or the name of a data source.
            remove_assets (bool): Set to true if Assets should be deleted as well.
        Returns:
            dict: Status object
        """
        body = {
            'deleteAssets': remove_assets
        }
        url = '/api/v1/data-sources/{}'.format(ds.id)
        return self.app.client.delete(url, body)



#!/usr/bin/env python3
from setuptools import setup

# See https://packaging.python.org/tutorials/packaging-projects/
# for details about packaging python projects

# Generating distribution archives (run from same directory as this file)
# python3 -m pip install --user --upgrade setuptools wheel
# python3 setup.py sdist bdist_wheel

requirements = [
    "requests",
    "pyOpenSSL",
    "PyJWT>=2.0",
    "backoff",
    "pytest"
]

setup(
    name='zvi-client',
    version="1.1.3",
    description='Zorroa Visual Intelligence Python Client',
    url='http://www.zorroa.com',
    license='Apache2',
    package_dir={'': 'pylib'},
    packages=['zmlp', 'zmlp.app', 'zmlp.entity'],
    scripts=[],
    author="Matthew Chambers",
    author_email="support@zorroa.com",
    keywords="machine learning artificial intelligence",
    python_requires='>=3.4',

    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],

    include_package_data=True,
    install_requires=requirements
)



from ..entity import Job, Task, TaskError
from ..util import as_collection, as_id_collection, as_id


class JobApp:
    """
    An App instance for managing Jobs. Jobs are containers for async processes
    such as data import or training.
    """
    def __init__(self, app):
        self.app = app

    def get_job(self, id):
        """
        Get a Job by its unique Id.
        Args:
            id (str): The Job id or Job object.

        Returns:
            Job: The Job
        """
        return Job(self.app.client.get('/api/v1/jobs/{}'.format(as_id(id))))

    def refresh_job(self, job):
        """
        Refreshes the internals of the given job.

        Args:
            job (Job): The job to refresh.

        """
        job._data = self.app.client.get('/api/v1/jobs/{}'.format(job.id))

    def find_jobs(self, id=None, state=None, name=None, limit=None, sort=None):
        """
        Find jobs matching the given criteria.

        Args:
            id (mixed): A job ID or IDs to filter on.
            state (mixed): A Job state or list of states to filter on.
            name (mixed): A Job name or list of names to filter on.
            limit (int): The maximum number of jobs to return, None  is no limit.
            sort (list): A list of sort ordering phrases, like ["name:d", "time_created:a"]

        Returns:
            generator: A generator which will return matching jobs when iterated.

        """
        body = {
            'ids': as_collection(id),
            'states': as_collection(state),
            'names': as_collection(name),
            'sort': sort
        }
        return self.app.client.iter_paged_results('/api/v1/jobs/_search', body, limit, Job)

    def find_one_job(self, id=None, state=None, name=None):
        """
        Find single Job matching the given criteria.  Raises exception if more
        than one result is found.

        Args:
            id (mixed): A job ID or IDs to filter on.
            state (mixed): A Job state or list of states to filter on.
            name (mixed): A Job name or list of names to filter on.
            sort (list): A list of sort ordering phrases, like ["name:d", "time_created:a"]

        Returns:
            Job: The job.
        """
        body = {
            'ids': as_collection(id),
            'states': as_collection(state),
            'names': as_collection(name)
        }
        return Job(self.app.client.post('/api/v1/jobs/_findOne', body))

    def find_task_errors(self, query=None, job=None, task=None,
                         asset=None, path=None, processor=None, limit=None, sort=None):
        """
        Find TaskErrors based on the supplied criterion.

        Args:
            query (str): keyword query to match various error properties.
            job (mixed): A single Job, job id or list of either type.
            task (mixed): A single Task, task id or list of either type.
            asset (mixed): A single Asset, asset id or list of either type.
            path (mixed): A file path or list of file path.
            processor (mixed): A processor name or list of processors.
            limit (int): Limit the number of results or None for all results.
            sort (list): A list of sort ordering phrases, like ["name:d", "time_created:a"]

        Returns:
            generator: A generator which returns results when iterated.

        """
        body = {
            'keywords': query,
            'jobIds': as_id_collection(job),
            'taskIds': as_id_collection(task),
            'assetIds': as_id_collection(asset),
            'paths': as_collection(path),
            'processor': as_collection(processor),
            'sort': sort
        }
        return self.app.client.iter_paged_results(
            '/api/v1/taskerrors/_search', body, limit, TaskError)

    def pause_job(self, job):
        """
        Pause scheduling for the given Job.  Pausing a job simply removes the
        job from scheduler consideration.  All existing tasks will continue to run
        and Analysts will move to new jobs as tasks complete.

        Args:
            job (Job): The Job to pause

        Returns:
            bool: True if the job was actually paused.

        """
        # Resolve the job if we need to.
        if isinstance(job, str):
            job = self.get_job(job)
        if self.app.client.put('/api/v1/jobs/{}'.format(job.id), job._data)['success']:
            job._data['paused'] = True
            return True
        return False

    def resume_job(self, job):
        """
        Resume scheduling for the given Job.

        Args:
            job (Job): The Job to resume

        Returns:
            bool: True of the job was actually resumed.
        """
        if isinstance(job, str):
            job = self.get_job(job)
        if self.app.client.put('/api/v1/jobs/{}'.format(job.id), job._data)['success']:
            job._data['paused'] = False
            return True
        return False

    def cancel_job(self, job):
        """
        Cancel the given Job.  Canceling a job immediately kills all running Tasks
        and removes the job from scheduler consideration.

        Args:
            job (Job): The Job to cancel, or the job's unique Id.

        Returns:
            bool: True if the job was actually canceled, False if the job was already cancelled.

        """
        if isinstance(job, str):
            job = self.get_job(job)
        if self.app.client.put('/api/v1/jobs/{}/_cancel'.format(job.id)).get('success'):
            self.refresh_job(job)
            return True
        return False

    def restart_job(self, job):
        """
        Restart a canceled job.

        Args:
            job (Job): The Job to restart

        Returns:
            bool: True if the job was actually restarted, false if the job was not cancelled.

        """
        if isinstance(job, str):
            job = self.get_job(job)
        if self.app.client.put('/api/v1/jobs/{}/_restart'.format(job.id)).get('success'):
            self.refresh_job(job)
            return True
        return False

    def retry_all_failed_tasks(self, job):
        """
        Retry all failed Tasks in the Job.

        Args:
            job (Job): The Job with failed tasks.

        Returns:
            bool: True if the some failed tasks were restarted.
        """
        if isinstance(job, str):
            job = self.get_job(job)
        if self.app.client.put(
                '/api/v1/jobs/{}/_retryAllFailures'.format(job.id)).get('success'):
            self.refresh_job(job)
            return True
        return False

    def find_tasks(self, job=None, id=None, name=None, state=None, limit=None, sort=None):
        """
        Find Tasks matching the given criteria.

        Args:
            job: (mixed): A single Job, job id or list of either type.
            id (mixed): A single Task, task id or list of either type.
            name (mixed): A task name or list of tasks names.
            state (mixed): A take state or list of task states.
            limit (int): Limit the number of results, None for no limit.
            sort (list): A list of sort ordering phrases, like ["name:d", "time_created:a"]

        Returns:
            generator: A Generator that returns matching Tasks when iterated.
        """
        body = {
            'ids': as_collection(id),
            'states': as_collection(state),
            'names': as_collection(name),
            'jobIds': as_id_collection(job),
            'sort': sort
        }
        return self.app.client.iter_paged_results('/api/v1/tasks/_search', body, limit, Task)

    def find_one_task(self, job=None, id=None, name=None, state=None):
        """
        Find a single task matching the criterion.

        Args:
            job: (mixed): A single Job, job id or list of either type.
            id (mixed): A single Task, task id or list of either type.
            name (mixed): A task name or list of tasks names.
            state (mixed): A take state or list of task states.

        Returns:
            Task A single matching task.

        """
        body = {
            'ids': as_collection(id),
            'states': as_collection(state),
            'names': as_collection(name),
            'jobIds': as_id_collection(job)
        }
        res = Task(self.app.client.post('/api/v1/tasks/_findOne', body))
        return res

    def get_task(self, task):
        """
        Get a Task by its unique id.

        Args:
            task (str): The Task or task id.

        Returns:
            Task: The Task

        """
        return Task(self.app.client.get('/api/v1/tasks/{}'.format(as_id(task))))

    def refresh_task(self, task):
        """
        Refreshes the internals of the given job.

        Args:
            task (Task): The Task

        """
        task._data = self.app.client.get('/api/v1/tasks/{}'.format(task.id))

    def skip_task(self, task):
        """
        Skip the given task.  A skipped task wilk not run.

        Args:
            task (str): The Task or task id.

        Returns:
            bool: True if the Task changed to the Skipped state.
        """
        if isinstance(task, str):
            task = self.get_task(task)
        if self.app.client.put('/api/v1/tasks/{}/_skip'.format(task.id))['success']:
            self.refresh_task(task)
            return True
        return False

    def retry_task(self, task):
        """
        Retry the given task.  Retried tasks are set back to the waiting state.

        Args:
            task (str): The Task or task id.

        Returns:
           bool: True if the Task changed to the Waiting state.
        """
        if isinstance(task, str):
            task = self.get_task(task)
        if self.app.client.put('/api/v1/tasks/{}/_retry'.format(task.id))['success']:
            self.refresh_task(task)
            return True
        return False

    def get_task_script(self, task):
        """
        Return the given task's ZPS script.

        Args:
            task: (str): The Task or task id.

        Returns:
            dict: The script in dictionary form.
        """
        return self.app.client.get('/api/v1/tasks/{}/_script'.format(as_id(task)))

    def download_task_log(self, task, dst_path):
        """
        Download the task log file to the given file path.

        Args:
            task: (str): The Task or task id.
            dst_path (str): The path to the destination file.
        Returns:
            dict: The script in dictionary form.
        """
        return self.app.client.stream('/api/v1/tasks/{}/_log'.format(as_id(task)), dst_path)

    def iterate_task_log(self, task):
        """
        Return a generator that can be used to iterate a task log file.

        Args:
            task: (str): The Task or task id.

        Returns:
            generator: A generator which yields each line of a log file.

        """
        return self.app.client.stream_text('/api/v1/tasks/{}/_log'.format(as_id(task)))



from ..entity.field import CustomField
from ..util import as_id


class CustomFieldApp:
    """
    An App instance for managing custom fields.
    """
    def __init__(self, app):
        self.app = app

    def create_custom_field(self, name, type):
        """
        Create a new Field.  The name of the field will be used as the actual
        ElasticSearch field name.  The name must be alpha-numeric, underscores/dashes
        are allowed, periods are not.

        To reference your custom field in an ES query you must prepend the field name
        with 'custom.'. For example if your field name is 'city', then you must use the
        fully qualified name 'custom.city' in your query.

        Args:
            name (str): The name of the field.
            type (str): The ES field type.

        Returns:
            CustomField: The new custom field.

        """
        body = {
            'name': name,
            'type': type
        }
        return CustomField(self.app.client.post('/api/v3/custom-fields', body))

    def get_custom_field(self, id):
        """
        Get the record for the custom field.

        Args:
            id (str): The id of the field.

        Returns:
             CustomField: The custom field.
        """
        id = as_id(id)
        return CustomField(self.app.client.get(f'/api/v3/custom-fields/{id}'))
